"""
@author: Yingjia Lin
"""

import warnings

import numpy as np
from scipy.sparse import csc_matrix
from tqdm import tqdm

from .base import spacetime


def sliding_window_phenom_mem(zcheck_samples, hz, lz, W, F, decoder1, decoder2, dict1: dict, dict2: dict, function_name1: str, function_name2: str, tqdm_on=False):
    '''
    Sliding window decoder in S. Huang and S. Puri, PRA 110, 012453 (2024)
    This version allows to replace the decoder used in the decoding scheme.
    The space-time code has not yet been implemented

    For convenience the notation assumes z-type memory, but the code works equivalently for x-type memory.

    :param zcheck_samples: 2-dim numpy array of detector results; see get_stim_Zmem_result in simulation.py. Shape (# trials, # Z-check qubits * (# rounds+1))
    :param hz: Parity check matrix (in the code-capacity level) representing the Z stabilizers of the qec code. Shape ((# Z-check qubits, # data qubits))
    :param lz: Logical codeword matrix of the qec code. Shape ((# logical qubits, # data qubits))
    :param W: Width of sliding window
    :param F: Width of overlap between consecutive sliding windows
    :param decoder1: A python class for an inner decoder for each window before the last window. This class can be initialized by some parameters and has a decode function that takes a syndrome and returns a correction.
    :param decoder2: A python class for an inner decoder for the last window including the ideal round of measurement
    :param dict1: parameters of decoder1, except for the parity check matrix
    :param dict2: parameters of decoder2, except for the parity check matrix
    :param function_name1: the decoding function name of decoder1 that is called to decode a syndrome after the initialization of the decoder1
    :param function_name2: the decoding function of decoder1 that is called to decode a syndrome after the initialization of the decoder2
    :param eff_error_rate_per_fault: Estimate of error rate in the context of code-capacity/phenomenological level simulations.
                For circuit level, we suggest p * (num_layers + 3), where p is the depolarizing error rate
                and num_layers is the circuit depth for each stabilizer measurement round
                e.g. for hgp codes, num_layers == code.count_color('east') + code.count_color('north') + code.count_color('south') + code.count_color('west')

    :return logical_z_pred: Decoder's prediction of whether the logical Z codewords flipped. Shape (# trials, # logical quits)
    '''
    if F == 0:
        raise ValueError("Input parameter F cannot be zero.")
    num_trials = zcheck_samples.shape[0]
    num_rounds = zcheck_samples.shape[1] // hz.shape[0] - 2

    # update the total number of windows for decoding, the size of the last window
    if 2 + num_rounds - W >= 0:
        num_cor_rounds = (2 + num_rounds - W) // F  # num_cor_rounds=num of windows before the last window
        if (2 + num_rounds - W) % F != 0:  # we can slide one more window if the remaining rounds>W
            num_cor_rounds += 1
    else:
        num_cor_rounds = 0
        warnings.warn("Window size larger than the syndrome extraction rounds: Doing whole history correction")
    W_last = num_rounds + 2 - F * num_cor_rounds

    # update the window matrix and the decoder
    B = np.eye(W, dtype=int)
    for i in range(1, W):
        B[i, i - 1] = 1
    hz_phenom = np.column_stack((np.kron(np.eye(W, dtype=int), hz), np.kron(B, np.eye(hz.shape[0], dtype=int))))

    decoder_each_window = decoder1(csc_matrix(hz_phenom), **dict1)

    B_last = np.eye(W_last, dtype=int)
    for i in range(1, W_last):
        B_last[i, i - 1] = 1
    B_last = B_last[:, :W_last - 1]  # The last round in this window is ideal

    hz_last = np.column_stack((np.kron(np.eye(W_last, dtype=int), hz), np.kron(B_last, np.eye(hz.shape[0], dtype=int))))
    decoder_last_window = decoder2(csc_matrix(hz_last), **dict2)

    if tqdm_on:
        iterator = tqdm(range(num_trials))
    else:
        iterator = range(num_trials)
    logical_z_pred = np.zeros((num_trials, lz.shape[0]), dtype=int)

    for i in iterator:  # each sample decoding
        accumulated_correction = np.zeros(hz.shape[1], dtype=int)
        syn_update = np.zeros(hz.shape[0], dtype=int)

        for k in range(num_cor_rounds):
            diff_syndrome = (zcheck_samples[i, F * k * hz.shape[0]:(F * k + W) * hz.shape[0]].copy()) % 2
            diff_syndrome[:hz.shape[0]] = (diff_syndrome[:hz.shape[0]] + syn_update) % 2

            decoded_errors = getattr(decoder_each_window, function_name1)(diff_syndrome)
            correction = np.sum(decoded_errors[:F * hz.shape[1]].reshape(F, hz.shape[1]).copy(), axis=0) % 2

            syn_update = decoded_errors[W * hz.shape[1] + (F - 1) * hz.shape[0]:W * hz.shape[1] + F * hz.shape[0]].copy()
            accumulated_correction = (accumulated_correction + correction) % 2

        # In the last round we just correct the whole window
        diff_syndrome = (zcheck_samples[i, (F * num_cor_rounds) * hz.shape[0]:].copy()) % 2
        diff_syndrome[:hz.shape[0]] = (diff_syndrome[:hz.shape[0]] + syn_update) % 2

        decoded_errors = getattr(decoder_last_window, function_name2)(diff_syndrome)
        correction = np.sum(decoded_errors[:W_last * hz.shape[1]].reshape(W_last, hz.shape[1]), axis=0) % 2

        accumulated_correction = (accumulated_correction + correction) % 2
        logical_z_pred[i, :] = (lz @ accumulated_correction) % 2

    return logical_z_pred


def sliding_window_circuit_mem(zcheck_samples, circuit, hz, lz, W, F, decoder1, decoder2, dict1: dict, dict2: dict,
                               error_rate_name1: str, error_rate_name2: str,
                               function_name1: str, function_name2: str, tqdm_on=False):
    '''
    Sliding window decoder in S. Huang and S. Puri, PRA 110, 012453 (2024) implemented with spacetime
    For convenience the notation assumes z-type memory, but the code works equivalently for x-type memory.

    :param zcheck_samples: 2-dim numpy array of detector results; see get_stim_Zmem_result in simulation.py. Shape (# trials, # Z-check qubits * (# rounds+1))
    :param circuit: simulated stim.Circuit
    :param hz: Parity check matrix (in the code-capacity level) representing the Z stabilizers of the qec code. Shape ((# Z-check qubits, # data qubits))
    :param lz: Logical codeword matrix of the qec code. Shape ((# logical qubits, # data qubits))
    :param W: Width of sliding window
    :param F: Width of overlap between consecutive sliding windows
    :param decoder1: A python class for an inner decoder for each window before the last window. This class can be initialized by some parameters and has a decode function that takes a syndrome and returns a correction.
    :param decoder2: A python class for an inner decoder for the last window including the ideal round of measurement
    :param dict1: parameters of decoder1, except for the parity check matrix
    :param dict2: parameters of decoder2, except for the parity check matrix
    :param error_rate_name1: the parameter name of the list of error rates for decoder1 for initializing decoder1
    :param error_rate_name2: the parameter name of the list of error rates for decoder2 for initializing decoder2
    :param function_name1: the decoding function name of decoder1 that is called to decode a syndrome after the initialization of the decoder1
    :param function_name2: the decoding function name of decoder2 that is called to decode a syndrome after the initialization of the decoder2
    :param tqdm_on: True or False: Evaluating the iteration runtime

    :return logical_z_pred: Decoder's prediction of whether the logical Z codewords flipped. Shape (# trials, # logical quits)
    '''

    num_trials = zcheck_samples.shape[0]
    num_rounds = zcheck_samples.shape[1] // hz.shape[0] - 2

    # update the total number of windows for decoding, the size of the last window
    if 2 + num_rounds - W >= 0:
        num_cor_rounds = (2 + num_rounds - W) // F  # num_cor_rounds=num of windows before the last window
        if (2 + num_rounds - W) % F != 0:  # we can slide one more window if the remaining rounds>W
            num_cor_rounds += 1
    else:
        num_cor_rounds = 0
        warnings.warn("Window size larger than the syndrome extraction rounds: Doing whole history correction")
    W_last = num_rounds + 2 - F * num_cor_rounds
    # update the window matrix and the decoder
    # spacetime detector error matrix
    window_check_set, window_observable_set, window_priors_set, window_update = spacetime(circuit, hz, W, F, num_cor_rounds)
    # decoder for each window
    decoder = []
    for i in range(len(window_check_set) - 1):
        dict1[error_rate_name1] = window_priors_set[i]
        decoder_each_window = decoder1(window_check_set[i], **dict1)
        decoder.append(decoder_each_window)
    dict2[error_rate_name2] = window_priors_set[len(window_check_set) - 1]
    decoder_each_window = decoder2(window_check_set[len(window_check_set) - 1], **dict2)
    decoder.append(decoder_each_window)

    # start decoding
    if tqdm_on:
        iterator = tqdm(range(num_trials))
    else:
        iterator = range(num_trials)
    logical_z_pred = np.zeros((num_trials, lz.shape[0]), dtype=int)

    for i in iterator:  # each sample decoding
        accumulated_correction = np.zeros(window_observable_set[0].shape[0], dtype=int)
        syn_update = np.zeros(hz.shape[0], dtype=int)

        for k in range(num_cor_rounds):
            # syndrome of the window
            diff_syndrome = (zcheck_samples[i, F * k * hz.shape[0]:(F * k + W) * hz.shape[0]].copy()) % 2
            diff_syndrome[:hz.shape[0]] = (diff_syndrome[:hz.shape[0]] + syn_update) % 2  # update the syndrome based on the previous window decoding

            decoded_errors = getattr(decoder[k], function_name1)(diff_syndrome)
            correction = window_observable_set[k] @ decoded_errors[:window_observable_set[k].shape[1]] % 2  # interpret the correction operation as final observable flips

            syn_update = window_update[k] @ decoded_errors[:window_observable_set[k].shape[1]] % 2
            accumulated_correction = (accumulated_correction + correction) % 2

        # In the last round we just correct the whole window
        # syndrome of last round
        diff_syndrome = (zcheck_samples[i, (F * num_cor_rounds) * hz.shape[0]:].copy()) % 2
        diff_syndrome[:hz.shape[0]] = (diff_syndrome[:hz.shape[0]] + syn_update) % 2
        # Observable flips based on correction
        decoded_errors = getattr(decoder[num_cor_rounds], function_name2)(diff_syndrome)
        correction = window_observable_set[num_cor_rounds] @ decoded_errors % 2
        accumulated_correction = (accumulated_correction + correction) % 2
        # Predicted observable flips
        logical_z_pred[i, :] = accumulated_correction

    return logical_z_pred


__all__ = ["sliding_window_phenom_mem", "sliding_window_circuit_mem"]



"""
@author: Mingyu Kang, Yingjia Lin
"""

import numpy as np

def get_qldpc_mem_circuit(code, idle_error, sqgate_error, tqgate_error, spam_error, num_rounds,\
                          basis='Z', get_all_detectors=False, noisy_init=True, noisy_meas=False):
    '''
    Returns the full Stim circuit for circuit-level simulations.
    Errors occur at each and every gate. 

    :param code: The qldpc_code class from qldpc.py
    :param idle_error: Rate of depolarizing error. Float or tuple of 3 floats (px, py, pz)
    :param sqgate_error: Rate of single-qubit gate error. Float or tuple of 3 floats (px, py, pz)
    :param tqgate_error: Rate of two-qubit gate error. Float or tuple of 15 floats (pix, ..., pzz)
    :param spam_error: Rate of state preparation and measurement error. Float
    :param num_rounds: Number of stabilizer measurement rounds
    :param basis: Basis in which logical codewords are stored. Options are 'Z' and 'X'.
    :param get_all_detectors: If True, all detectors in the Z and X bases are obtained. 
                              If False, only the detectors in the memory basis are obtained. 
    :param noisy_init: If False, reset of data and syndrome qubits at the 0-th round is assumed to be error free.
    :param noisy_meas: If False, final measurement of data qubits is assumed to be error free
    :return circuit: String that can be converted to Stim circuit by stim.Circuit()
    '''
    
    basis = basis.upper()
    get_Z_detectors = True if basis == 'Z' or get_all_detectors else False
    get_X_detectors = True if basis == 'X' or get_all_detectors else False
    directions = list(code.direction_inds.keys())
    circ = Circuit(code.all_qubits)
    
    def _add_stabilizer_round():
        circ.add_hadamard_layer(code.xcheck_qubits)
        for direction_ind in range(len(directions)):
            direction = directions[direction_ind]
            for color in range(code.num_colors[direction]):
                edges = code.colored_edges[direction_ind][color]
                circ.add_cnot_layer(edges)
        circ.add_hadamard_layer(code.xcheck_qubits)
        circ.add_measure_reset_layer(code.check_qubits)
        return
    
    ################## Logical state prep ##################
    if noisy_init:
        circ.set_error_rates(idle_error, sqgate_error, tqgate_error, spam_error)
    else:
        circ.set_error_rates(0., 0., 0., 0.)
    circ.add_reset(code.data_qubits, basis)
    circ.add_reset(code.check_qubits)
    circ.add_tick()

    circ.set_error_rates(idle_error, sqgate_error, tqgate_error, spam_error)
    _add_stabilizer_round()

    if basis == 'Z':
        for i in range(1, len(code.zcheck_qubits)+1)[::-1]:     
            circ.add_detector([len(code.xcheck_qubits) + i])
    elif basis == 'X':
        for i in range(1, len(code.xcheck_qubits)+1)[::-1]:
            circ.add_detector([i])   
    
    ############## Logical memory w/ noise ###############
    circ.set_error_rates(idle_error, sqgate_error, tqgate_error, spam_error)
    
    if num_rounds > 0: 
        circ.start_loop(num_rounds)  
        
        _add_stabilizer_round()
        
        if get_Z_detectors:
            for i in range(1, len(code.zcheck_qubits)+1)[::-1]:
                ind = len(code.xcheck_qubits) + i        
                circ.add_detector([ind, ind + len(code.check_qubits)])
        if get_X_detectors:
            for i in range(1, len(code.xcheck_qubits)+1)[::-1]:
                circ.add_detector([i, i + len(code.check_qubits)])
            
        circ.end_loop()
        
    ################## Logical measurement ##################   
    if not noisy_meas:
        circ.set_error_rates(0., 0., 0., 0.)
        
    circ.add_measure(code.data_qubits, basis)
    
    if basis == 'Z':
        for i in range(1, len(code.zcheck_qubits)+1)[::-1]:
            inds = np.array([len(code.data_qubits) + len(code.xcheck_qubits) + i])
            inds = np.concatenate((inds, len(code.data_qubits) - np.where(code.hz[len(code.zcheck_qubits)-i, :]==1)[0]))
            circ.add_detector(inds)
        
        for i in range(len(code.lz)):
            circ.add_observable(i, len(code.data_qubits) - np.where(code.lz[i,:]==1)[0])
            
    elif basis == 'X':
        for i in range(1, len(code.xcheck_qubits)+1)[::-1]:
            inds = np.array([len(code.data_qubits) + i])
            inds = np.concatenate((inds, len(code.data_qubits) - np.where(code.hx[len(code.xcheck_qubits)-i, :]==1)[0]))
            circ.add_detector(inds)
        
        for i in range(len(code.lx)):
            circ.add_observable(i, len(code.data_qubits) - np.where(code.lx[i,:]==1)[0])     
            
    return circ.circuit

def check_overlapping_CX(circuit, verbose=True):
    '''
    Check for overlapping CX gates in the same layer of a stim circuit.

    :param circuit: stim.Circuit
    :param verbose: If True, print overlaps as they are found.
    :return: List of (index, duplicates) for each offending CX instruction.
    '''
    overlaps = []
    for i in range(len(circuit)):
        if circuit[i].name == 'CX':
            size = len(circuit[i].targets_copy())
            gate_list = np.zeros(size, dtype=int)
            for j in range(size):
                gate_list[j] = circuit[i].targets_copy()[j].qubit_value
            unique, counts = np.unique(gate_list, return_counts=True)

            duplicates = unique[counts > 1]
            if duplicates.size > 0:
                if verbose:
                    print("Duplicates found:", i, duplicates)
                overlaps.append((i, duplicates.copy()))
    if verbose and not overlaps:
        print("No overlapping CX gates found.")
    return overlaps


class Circuit:
    '''
    Class containing helper functions for writing Stim circuits (https://github.com/quantumlib/Stim)
    '''
    
    def __init__(self, all_qubits):
        
        self.circuit = ''
        self.margin = ''
        self.all_qubits = all_qubits
        self.idle_error = 0.
        self.sqgate_error = 0.
        self.tqgate_error = 0.
        self.spam_error = 0.
        
    def set_all_qubits(self, all_qubits):
        self.all_qubits = all_qubits
    
    def set_error_rates(self, idle_error, sqgate_error, tqgate_error, spam_error):
        self.idle_error = idle_error
        self.sqgate_error = sqgate_error
        self.tqgate_error = tqgate_error
        self.spam_error = spam_error       
        
    def start_loop(self, num_rounds):
        c = 'REPEAT %d {\n'%num_rounds
        self.circuit += c
        self.margin = '    ' 
        return c
        
    def end_loop(self):
        c = '}\n'
        self.circuit += c
        self.margin = ''
        return c
        
    def add_tick(self):
        c = self.margin + 'TICK\n'
        self.circuit += c
        return c   
        
    def add_reset(self, qubits, basis='Z'):        
        basis = basis.upper()
        
        c = self.margin
        if basis == 'Z':
            c += 'R '
        elif basis == 'X':
            c += 'RX '
        for q in qubits:
            c += '%d '%q
        c += '\n'
        
        if self.spam_error > 0.:
            c += self.margin
            if basis == 'Z':
                c += 'X_ERROR(%.10f) '%self.spam_error
            elif basis == 'X':
                c += 'Z_ERROR(%.10f) '%self.spam_error            
            for q in qubits:
                c += '%d '%q            
            c += '\n'
        
        self.circuit += c
        return c       
    
    def add_idle(self, qubits):
        if self.idle_error == 0.:
            return ''
        
        c = self.margin
        if type(self.idle_error) == float or type(self.idle_error) == np.float64:
            c += 'DEPOLARIZE1(%.10f) '%self.idle_error
        else:
            c += 'PAULI_CHANNEL_1('
            for p in self.idle_error:
                c += '%.10f, '%p
            c = c[:-2] + ') '
        for q in qubits:
            c += '%d '%q
        c += '\n'
        
        self.circuit += c
        return c
    
    def add_hadamard(self, qubits):
        c = self.margin
        c += 'H '
        for q in qubits:
            c += '%d '%q
        c += '\n'
        
        if self.sqgate_error != 0. and self.sqgate_error != 0:
            c += self.margin
            if type(self.sqgate_error) == float or type(self.sqgate_error) == np.float64:
                c += 'DEPOLARIZE1(%.10f) '%self.sqgate_error
            else:
                c += 'PAULI_CHANNEL_1('
                for p in self.sqgate_error:
                    c += '%.10f, '%p
                c = c[:-2] + ') '
            for q in qubits:
                c += '%d '%q
            c += '\n'
            
        self.circuit += c
        return c
    
    def add_hadamard_layer(self, qubits):
        c1 = self.add_hadamard(qubits)
        other_qubits = np.delete(self.all_qubits, np.where(np.isin(self.all_qubits, qubits))[0])
        c2 = self.add_idle(other_qubits)
        c3 = self.add_tick()
        return c1 + c2 + c3
    
    def add_cnot(self, qubits):
        c = self.margin
        c += 'CX '
        for q in qubits:
            c += '%d '%q
        c += '\n'
        
        if self.tqgate_error != 0. and self.tqgate_error != 0:
            c += self.margin
            if type(self.tqgate_error) == float or type(self.tqgate_error) == np.float64:
                c += 'DEPOLARIZE2(%.10f) '%self.tqgate_error
            else:
                c += 'PAULI_CHANNEL_2('
                for p in self.tqgate_error:
                    c += '%.10f, '%p
                c = c[:-2] + ') '
            for q in qubits:
                c += '%d '%q
            c += '\n'
            
        self.circuit += c
        return c        
        
    def add_cnot_layer(self, qubits):
        c1 = self.add_cnot(qubits)
        other_qubits = np.delete(self.all_qubits, np.where(np.isin(self.all_qubits, qubits))[0])
        c2 = self.add_idle(other_qubits)
        c3 = self.add_tick()
        return c1 + c2 + c3    
    
    def add_measure_reset(self, qubits, error_free_reset=False):       
        c = ''
        if self.spam_error > 0.:
            c += self.margin
            c += 'X_ERROR(%.10f) '%self.spam_error           
            for q in qubits:
                c += '%d '%q            
            c += '\n'
            
        c += self.margin
        c += 'MR '
        for q in qubits:
            c += '%d '%q
        c += '\n'   
        
        if self.spam_error > 0. and not error_free_reset:
            c += self.margin
            c += 'X_ERROR(%.10f) '%self.spam_error          
            for q in qubits:
                c += '%d '%q            
            c += '\n'        
            
        self.circuit += c
        return c
    
    def add_measure_reset_layer(self, qubits, error_free_reset=False):
        c1 = self.add_measure_reset(qubits, error_free_reset)
        other_qubits = np.delete(self.all_qubits, np.where(np.isin(self.all_qubits, qubits))[0])
        c2 = self.add_idle(other_qubits)
        c3 = self.add_tick()
        return c1 + c2 + c3  
        
    def add_measure(self, qubits, basis='Z'):
        basis = basis.upper()
        
        c = ''
        if self.spam_error > 0.:
            c += self.margin
            if basis == 'Z':
                c += 'X_ERROR(%.10f) '%self.spam_error
            elif basis == 'X':
                c += 'Z_ERROR(%.10f) '%self.spam_error            
            for q in qubits:
                c += '%d '%q            
            c += '\n'
            
        c += self.margin
        if basis == 'Z':
            c += 'M '
        elif basis == 'X':
            c += 'MX '
        for q in qubits:
            c += '%d '%q
        c += '\n'        
        
        self.circuit += c
        return c 
    
    def add_detector(self, inds):
        c = self.margin + 'DETECTOR '
        for ind in inds:
            c += 'rec[-%d] '%ind
        c += '\n'
        
        self.circuit += c
        
    def add_observable(self, observable_no, inds):
        c = self.margin + 'OBSERVABLE_INCLUDE(%d) '%observable_no
        for ind in inds:
            c += 'rec[-%d] '%ind
        c += '\n'
        
        self.circuit += c
        return c
    



"""
@author: Mingyu Kang, Yingjia Lin
"""

import numpy as np

from .circuit_construction import get_builder
from ..gf2_util import compute_lz_and_lx
from .base import QldpcCode


class QlpCode(QldpcCode):
    def __init__(self, b1, b2, lift_size):
        '''
        :param b1: First base matrix used to construct the lp code. Each entry is the power of the monomial.
                   e.g. b1 = np.array([[0, 0], [0, 3]]) represents the matrix of monomials [[1, 1], [1, x^3]].
        :param b2: Second base matrix used to construct the lp code. Each entry is the power of the monomial.
        :param lift_size: Size of cyclic matrix to which each monomial entry is lifted.
        '''
        # Reference base matrices: Q. Xu et al., arXiv:2308.08648 (quasi-cyclic lifted product codes).
        super().__init__()

        self.b1, self.b2 = b1, b2
        self.lift_size = lift_size
        self.m1, self.n1 = b1.shape
        self.m2, self.n2 = b2.shape

        b1T = (self.lift_size - b1).T % self.lift_size
        b2T = (self.lift_size - b2).T % self.lift_size
        b1_placeholder = np.ones(b1.shape, dtype=int)
        b2_placeholder = np.ones(b2.shape, dtype=int)

        hz_base = np.concatenate((np.kron(b2, np.eye(self.n1, dtype=int)),
                                  np.kron(np.eye(self.m2, dtype=int), b1T)), axis=1)
        hx_base = np.concatenate((np.kron(np.eye(self.n2, dtype=int), b1),
                                  np.kron(b2T, np.eye(self.m1, dtype=int))), axis=1)
        hz_base_placeholder = np.concatenate((np.kron(b2_placeholder, np.eye(self.n1, dtype=int)),
                                              np.kron(np.eye(self.m2, dtype=int), b1_placeholder.T)), axis=1)
        hx_base_placeholder = np.concatenate((np.kron(np.eye(self.n2, dtype=int), b1_placeholder),
                                              np.kron(b2_placeholder.T, np.eye(self.m1, dtype=int))), axis=1)

        self.hz = self.lift(self.lift_size, hz_base, hz_base_placeholder)
        self.hx = self.lift(self.lift_size, hx_base, hx_base_placeholder)
        self.lz, self.lx = compute_lz_and_lx(self.hz, self.hx)

    def build_circuit(self, strategy="cardinal", **opts):
        if strategy != "cardinal":
            return super().build_circuit(strategy=strategy, **opts)
        return self._build_cardinal_graph(**opts)

    def _build_cardinal_graph(self, seed=1):
        get_builder("cardinal").build_graph(self)
        data_qubits, zcheck_qubits, xcheck_qubits = [], [], []

        # Add nodes to the Tanner graph
        for i in range(self.n1):
            for j in range(self.n2):
                for l in range(self.lift_size):
                    node = (i + j * (self.n1 + self.m1)) * self.lift_size + l
                    data_qubits += [node]
                    self.graph.add_node(node, pos=(i, j))
                    self.node_colors += ['blue']

        start = self.n1 * self.lift_size
        for i in range(self.m1):
            for j in range(self.n2):
                for l in range(self.lift_size):
                    node = start + (i + j * (self.n1 + self.m1)) * self.lift_size + l
                    xcheck_qubits += [node]
                    self.graph.add_node(node, pos=(i + self.n1, j))
                    self.node_colors += ['purple']

        start = self.n2 * (self.n1 + self.m1) * self.lift_size
        for i in range(self.n1):
            for j in range(self.m2):
                for l in range(self.lift_size):
                    node = start + (i + j * (self.n1 + self.m1)) * self.lift_size + l
                    zcheck_qubits += [node]
                    self.graph.add_node(node, pos=(i, j + self.n2))
                    self.node_colors += ['green']

        start = (self.n2 * (self.n1 + self.m1) + self.n1) * self.lift_size
        for i in range(self.m1):
            for j in range(self.m2):
                for l in range(self.lift_size):
                    node = start + (i + j * (self.n1 + self.m1)) * self.lift_size + l
                    data_qubits += [node]
                    self.graph.add_node(node, pos=(i + self.n1, j + self.n2))
                    self.node_colors += ['blue']

        self.data_qubits = sorted(np.array(data_qubits))
        self.zcheck_qubits = sorted(np.array(zcheck_qubits))
        self.xcheck_qubits = sorted(np.array(xcheck_qubits))
        self.check_qubits = np.concatenate((self.zcheck_qubits, self.xcheck_qubits))
        self.all_qubits = sorted(np.array(data_qubits + zcheck_qubits + xcheck_qubits))

        hedge_bool_list = self.get_classical_edge_bools(np.ones(self.b1.shape, dtype=int), seed)
        vedge_bool_list = self.get_classical_edge_bools(np.ones(self.b2.shape, dtype=int), seed)

        edge_no = 0
        for i in range(self.m1):
            for j in range(self.n1):
                shift = self.b1[i, j]
                edge_bool = hedge_bool_list[(i, j)]

                for l in range(self.lift_size):
                    for k in range(self.n2 + self.m2):
                        if (k < self.n2) ^ edge_bool:
                            direction_ind = self.direction_inds['E']
                        else:
                            direction_ind = self.direction_inds['W']

                        control = (k * (self.n1 + self.m1) + self.n1 + i) * self.lift_size + (l + shift) % self.lift_size
                        target = (k * (self.n1 + self.m1) + j) * self.lift_size + l
                        self.add_edge(edge_no, direction_ind, control, target)
                        edge_no += 1

        for i in range(self.m2):
            for j in range(self.n2):
                shift = self.b2[i, j]
                edge_bool = vedge_bool_list[(i, j)]

                for l in range(self.lift_size):
                    for k in range(self.n1 + self.m1):
                        if (k < self.n1) ^ edge_bool:
                            direction_ind = self.direction_inds['N']
                        else:
                            direction_ind = self.direction_inds['S']

                        control = (k + j * (self.n1 + self.m1)) * self.lift_size + l
                        target = (k + (i + self.n2) * (self.n1 + self.m1)) * self.lift_size + (l + shift) % self.lift_size
                        self.add_edge(edge_no, direction_ind, control, target)
                        edge_no += 1

        # Color the edges of self.graph
        self.color_edges()
        return


class QlpPolyCode(QldpcCode):
    def __init__(self, b1, b2, lift_size):
        '''
        :param b1: First base matrix used to construct the lp code. Each entry is the list of powers of the polynomial terms.
                   e.g. b1 = [[[0], [0,1], []], [[], [0], [0,1]]] represents the matrix of monomials [[1, 1+x, 0], [0, 1, 1+x]].
        :param b2: Second base matrix used to construct the lp code. Each entry is the list of powers of the polynomial terms.
        :param lift_size: Size of cyclic matrix to which each polynomial term is lifted.
        '''
        # Polynomial-entry variant used for lift-connected surface codes; see Old, Rispler, Muller (arXiv:2401.02911).
        super().__init__()

        self.b1, self.b2 = b1, b2
        self.lift_size = lift_size

        self.m1, self.n1 = len(b1), len(b1[0])
        self.m2, self.n2 = len(b2), len(b2[0])

        # Base matrices where each entry ENCODEs the powers of polynomial terms in base of lift_size
        b1_enc = np.zeros((self.m1, self.n1), dtype=int)
        b1T_enc = np.zeros((self.n1, self.m1), dtype=int)
        b2_enc = np.zeros((self.m2, self.n2), dtype=int)
        b2T_enc = np.zeros((self.n2, self.m2), dtype=int)
        self.b1_placeholder = np.zeros((self.m1, self.n1), dtype=int)
        self.b2_placeholder = np.zeros((self.m2, self.n2), dtype=int)

        for i in range(self.m1):
            for j in range(self.n1):
                if self.b1[i][j] == []:
                    continue
                self.b1_placeholder[i, j] = 1

                bij, bTij = 0, 0
                for k in range(len(self.b1[i][j])):
                    bij += self.lift_size ** k * self.b1[i][j][k]
                    bTij += self.lift_size ** k * ((self.lift_size - self.b1[i][j][k]) % self.lift_size)
                b1_enc[i, j] = bij
                b1T_enc[j, i] = bTij

        for i in range(self.m2):
            for j in range(self.n2):
                if self.b2[i][j] == []:
                    continue
                self.b2_placeholder[i, j] = 1

                bij, bTij = 0, 0
                for k in range(len(self.b2[i][j])):
                    bij += self.lift_size ** k * self.b2[i][j][k]
                    bTij += self.lift_size ** k * ((self.lift_size - self.b2[i][j][k]) % self.lift_size)
                b2_enc[i, j] = bij
                b2T_enc[j, i] = bTij

        hz_base_enc = np.concatenate((np.kron(b2_enc, np.eye(self.n1, dtype=int)),
                                      np.kron(np.eye(self.m2, dtype=int), b1T_enc)), axis=1)
        hx_base_enc = np.concatenate((np.kron(np.eye(self.n2, dtype=int), b1_enc),
                                      np.kron(b2T_enc, np.eye(self.m1, dtype=int))), axis=1)
        hz_base_placeholder = np.concatenate((np.kron(self.b2_placeholder, np.eye(self.n1, dtype=int)),
                                              np.kron(np.eye(self.m2, dtype=int), self.b1_placeholder.T)), axis=1)
        hx_base_placeholder = np.concatenate((np.kron(np.eye(self.n2, dtype=int), self.b1_placeholder),
                                              np.kron(self.b2_placeholder.T, np.eye(self.m1, dtype=int))), axis=1)

        self.hz = self.lift_enc(self.lift_size, hz_base_enc, hz_base_placeholder)
        self.hx = self.lift_enc(self.lift_size, hx_base_enc, hx_base_placeholder)
        self.lz, self.lx = compute_lz_and_lx(self.hz, self.hx)

    def build_circuit(self, strategy="cardinal", **opts):
        if strategy != "cardinal":
            return super().build_circuit(strategy=strategy, **opts)
        return self._build_cardinal_graph(**opts)

    def _build_cardinal_graph(self, seed=1):
        get_builder("cardinal").build_graph(self)
        data_qubits, zcheck_qubits, xcheck_qubits = [], [], []

        # Add nodes to the Tanner graph
        for i in range(self.n1):
            for j in range(self.n2):
                for l in range(self.lift_size):
                    node = (i + j * (self.n1 + self.m1)) * self.lift_size + l
                    data_qubits += [node]
                    self.graph.add_node(node, pos=(i, j))
                    self.node_colors += ['blue']

        start = self.n1 * self.lift_size
        for i in range(self.m1):
            for j in range(self.n2):
                for l in range(self.lift_size):
                    node = start + (i + j * (self.n1 + self.m1)) * self.lift_size + l
                    xcheck_qubits += [node]
                    self.graph.add_node(node, pos=(i + self.n1, j))
                    self.node_colors += ['purple']

        start = self.n2 * (self.n1 + self.m1) * self.lift_size
        for i in range(self.n1):
            for j in range(self.m2):
                for l in range(self.lift_size):
                    node = start + (i + j * (self.n1 + self.m1)) * self.lift_size + l
                    zcheck_qubits += [node]
                    self.graph.add_node(node, pos=(i, j + self.n2))
                    self.node_colors += ['green']

        start = (self.n2 * (self.n1 + self.m1) + self.n1) * self.lift_size
        for i in range(self.m1):
            for j in range(self.m2):
                for l in range(self.lift_size):
                    node = start + (i + j * (self.n1 + self.m1)) * self.lift_size + l
                    data_qubits += [node]
                    self.graph.add_node(node, pos=(i + self.n1, j + self.n2))
                    self.node_colors += ['blue']

        self.data_qubits = sorted(np.array(data_qubits))
        self.zcheck_qubits = sorted(np.array(zcheck_qubits))
        self.xcheck_qubits = sorted(np.array(xcheck_qubits))
        self.check_qubits = np.concatenate((self.zcheck_qubits, self.xcheck_qubits))
        self.all_qubits = sorted(np.array(data_qubits + zcheck_qubits + xcheck_qubits))

        hedge_bool_list = self.get_classical_edge_bools(self.b1_placeholder, seed)
        vedge_bool_list = self.get_classical_edge_bools(self.b2_placeholder, seed)

        edge_no = 0
        for i in range(self.m1):
            for j in range(self.n1):
                if self.b1_placeholder[i, j] == 0:
                    continue
                edge_bool = hedge_bool_list[(i, j)]

                for l in range(self.lift_size):
                    for k in range(self.n2 + self.m2):
                        if (k < self.n2) ^ edge_bool:
                            direction_ind = self.direction_inds['E']
                        else:
                            direction_ind = self.direction_inds['W']

                        for shift in self.b1[i][j]:
                            control = (k * (self.n1 + self.m1) + self.n1 + i) * self.lift_size + (l + shift) % self.lift_size
                            target = (k * (self.n1 + self.m1) + j) * self.lift_size + l
                            self.add_edge(edge_no, direction_ind, control, target)
                            edge_no += 1

        for i in range(self.m2):
            for j in range(self.n2):
                if self.b2_placeholder[i, j] == 0:
                    continue
                edge_bool = vedge_bool_list[(i, j)]

                for l in range(self.lift_size):
                    for k in range(self.n1 + self.m1):
                        if (k < self.n1) ^ edge_bool:
                            direction_ind = self.direction_inds['N']
                        else:
                            direction_ind = self.direction_inds['S']

                        for shift in self.b2[i][j]:
                            control = (k + j * (self.n1 + self.m1)) * self.lift_size + l
                            target = (k + (i + self.n2) * (self.n1 + self.m1)) * self.lift_size + (l + shift) % self.lift_size
                            self.add_edge(edge_no, direction_ind, control, target)
                            edge_no += 1

        # Color the edges of self.graph
        self.color_edges()
        return


__all__ = ["QlpCode", "QlpPolyCode"]



import numpy as np
import stim
from ldpc.bposd_decoder import BpOsdDecoder

from quits.circuit import get_qldpc_mem_circuit
from quits.decoder.sliding_window import sliding_window_circuit_mem, sliding_window_phenom_mem
from quits.qldpc_code import HgpCode
from quits.simulation import get_stim_mem_result


def _build_hgp_code():
    h = np.loadtxt(
        "parity_check_matrices/n=12_dv=3_dc=4_dist=6.txt",
        dtype=int,
    )
    code = HgpCode(h, h)
    code.build_circuit(strategy="cardinal", seed=22)
    return code


def _simulate_mem_circuit(code, p, num_rounds, num_trials, basis="Z"):
    circuit = stim.Circuit(
        get_qldpc_mem_circuit(
            code,
            p,
            p,
            p,
            p,
            num_rounds,
            basis=basis,
        )
    )
    detection_events, observable_flips = get_stim_mem_result(
        circuit,
        num_trials,
        seed=1,
    )
    return circuit, detection_events, observable_flips


def _bp_osd_params(max_iter, osd_order):
    return {
        "bp_method": "product_sum",
        "max_iter": max_iter,
        "schedule": "serial",
        "osd_method": "osd_cs",
        "osd_order": osd_order,
    }


def test_sliding_window_circuit_mem_low_logical_error():
    code = _build_hgp_code()
    p = 1e-3
    num_rounds = 15
    num_trials = 50
    W, F = 5, 3
    max_iter, osd_order = 10, 1

    circuit, detection_events, observable_flips = _simulate_mem_circuit(
        code,
        p,
        num_rounds,
        num_trials,
    )

    dict1 = _bp_osd_params(max_iter, osd_order)
    dict2 = _bp_osd_params(max_iter, osd_order)

    logical_pred = sliding_window_circuit_mem(
        detection_events,
        circuit,
        code.hz,
        code.lz,
        W,
        F,
        BpOsdDecoder,
        BpOsdDecoder,
        dict1,
        dict2,
        "channel_probs",
        "channel_probs",
        "decode",
        "decode",
        tqdm_on=False,
    )

    pL = np.mean((observable_flips - logical_pred).any(axis=1))
    lfr = 1 - (1 - pL) ** (1 / num_rounds)

    print(
        "sliding_window_circuit_mem",
        {
            "p": p,
            "num_rounds": num_rounds,
            "num_trials": num_trials,
            "W": W,
            "F": F,
            "max_iter": max_iter,
            "osd_order": osd_order,
            "pL": float(pL),
            "lfr": float(lfr),
        },
    )
    print()

    assert pL <= 0.2
    assert lfr <= 0.05


def test_sliding_window_phenom_mem_low_logical_error():
    code = _build_hgp_code()
    p = 1e-3
    num_rounds = 15
    num_trials = 50
    W, F = 5, 3
    max_iter, osd_order = 10, 1

    circuit, detection_events, observable_flips = _simulate_mem_circuit(
        code,
        p,
        num_rounds,
        num_trials,
    )

    depth = sum(code.num_colors.values())
    eff_error_rate_per_fault = p * (depth + 3)

    dict1 = _bp_osd_params(max_iter, osd_order)
    dict2 = _bp_osd_params(max_iter, osd_order)
    dict1["error_rate"] = float(eff_error_rate_per_fault)
    dict2["error_rate"] = float(eff_error_rate_per_fault)

    logical_pred = sliding_window_phenom_mem(
        detection_events,
        code.hz,
        code.lz,
        W,
        F,
        BpOsdDecoder,
        BpOsdDecoder,
        dict1,
        dict2,
        "decode",
        "decode",
        tqdm_on=False,
    )

    pL = np.mean((observable_flips - logical_pred).any(axis=1))
    lfr = 1 - (1 - pL) ** (1 / num_rounds)

    print(
        "sliding_window_phenom_mem",
        {
            "p": p,
            "num_rounds": num_rounds,
            "num_trials": num_trials,
            "W": W,
            "F": F,
            "max_iter": max_iter,
            "osd_order": osd_order,
            "depth": depth,
            "eff_error_rate_per_fault": float(eff_error_rate_per_fault),
            "pL": float(pL),
            "lfr": float(lfr),
        },
    )

    assert pL <= 0.25
    assert lfr <= 0.08



"""
@author: Hanwen Yao, Mert Gökduman
LDPC utilities focused on:
  - Generating classical LDPC parity-check matrices
  - Computing / optimizing girth of the Tanner graph
"""

import numpy as np
import random
from collections import deque
from typing import Optional, Tuple
from .gf2_util import gf2_rank


# ============================================================
# LDPC generation (configuration model; may include multi-edges)
# ============================================================

def generate_ldpc(n: int, dv: int, dc: int) -> np.ndarray:
    """
    Generate an (m x n) LDPC parity-check matrix using a configuration model.
    Note: entries may exceed 1 (multi-edges).
    """
    if (n * dv) % dc != 0:
        raise ValueError("n * dv must be divisible by dc")
    m = (n * dv) // dc  # number of rows (check nodes)

    # Create a list with dv copies of each column index.
    col_sockets = []
    for col in range(n):
        col_sockets.extend([col] * dv)

    # Create a list with dc copies of each row index.
    row_sockets = []
    for row in range(m):
        row_sockets.extend([row] * dc)

    # There are the same total number of sockets in each list.
    assert len(col_sockets) == len(row_sockets)

    # Shuffle the row sockets and pair them with column sockets.
    random.shuffle(row_sockets)
    H = np.zeros((m, n), dtype=int)
    for col, row in zip(col_sockets, row_sockets):
        H[row, col] += 1    # Increment the entry to count multiple edges 
    return H


def has_duplicate_edges(H: np.ndarray) -> bool:
    """
    Returns True if the binary matrix H has any entry greater than 1,
    indicating the presence of multiple edges.
    """
    return bool(np.any(H > 1))


# ============================================================
# Girth computation on Tanner graph (binary support of H)
# ============================================================

def build_bipartite_adjacency(H: np.ndarray):
    """
    Build adjacency list for the bipartite graph represented by parity-check matrix H.
    Rows -> check nodes, columns -> bit nodes.
    """
    H = (np.asarray(H) > 0).astype(np.uint8)
    M, N = H.shape
    adjacency_list = [[] for _ in range(M + N)]
    for i in range(M):
        cols = np.where(H[i, :] == 1)[0]
        for j in cols:
            # Add an edge i <-> (M + j)
            adjacency_list[i].append(M + j)
            adjacency_list[M + j].append(i)
    return adjacency_list


def bfs_shortest_cycle(adjacency_list, start: int) -> int:
    """
    Performs BFS from 'start' node to find the shortest cycle reachable
    from this node. Returns the length of that cycle or infinity if none found.
    """
    dist = [-1] * len(adjacency_list)   # distance array
    dist[start] = 0
    q = deque([start])
    min_cycle_len = float("inf")

    while q:
        cur = q.popleft()
        for nb in adjacency_list[cur]:
            if dist[nb] == -1:
                # If neighbor is unvisited, set distance and enqueue
                dist[nb] = dist[cur] + 1
                q.append(nb)
            else:
                # If neighbor is visited (dist[neighbor] != -1)
                # and it's not the immediate parent of current in BFS tree,
                # then we have found a cycle.
                #
                # In an undirected BFS, the immediate parent of 'current' 
                # is the node with dist[current] - 1, i.e., one less distance.
                # A typical check is: if dist[neighbor] >= dist[current], 
                # we found a cycle that is not just an edge back to the parent.
                if dist[nb] >= dist[cur]:
                    cycle_len = dist[nb] + dist[cur] + 1
                    if cycle_len < min_cycle_len:
                        min_cycle_len = cycle_len

    return min_cycle_len


def girth_of_bipartite(adjacency_list) -> int:
    """
    Computes the girth (length of shortest cycle) of the bipartite graph
    represented by 'adjacency_list'.
    Returns float('inf') if there is no cycle.
    """
    best = float("inf")
    for node in range(len(adjacency_list)):
        best = min(best, bfs_shortest_cycle(adjacency_list, node))
    return best


def compute_girth_from_parity_check(H: np.ndarray) -> int:
    """
    Given a parity-check matrix H (M x N),
    build its bipartite adjacency list, then compute and return the girth.
    """
    return girth_of_bipartite(build_bipartite_adjacency(H))


# ============================================================
# Girth-oriented optimization (edge swapping on multi-edge Tanner)
# ============================================================

def get_neighbors(node, H):
    """
    Given a node in the Tanner graph (represented as a tuple):
      - ('v', i) for a variable node i, or
      - ('c', j) for a check node j,
    return a list of (neighbor, mult) pairs.
    Here, mult is the multiplicity (i.e. the count from H).
    """
    neighbors = []
    if node[0] == "v":
        i = node[1]
        # Neighbors are check nodes j with H[j, i] > 0.
        for j in range(H.shape[0]):
            if H[j, i] > 0:
                neighbors.append((("c", j), H[j, i]))
    else:
        j = node[1]
        # Neighbors are variable nodes i with H[j, i] > 0.
        for i in range(H.shape[1]):
            if H[j, i] > 0:
                neighbors.append((("v", i), H[j, i]))
    return neighbors


def dfs_count_paths(current, target, H, v_exclude, visited, depth, max_depth):
    """
    Recursively count the number of simple paths from 'current' to 'target'
    in the Tanner graph (given by H) with depth at most max_depth.
    We multiply counts by the edge multiplicities.
    
    v_exclude is the index of the variable node (v) for which we are counting cycles.
    We do not allow v_exclude as an intermediate node (except when it is the target).
    
    visited is a set of nodes (tuples) already on the current path (to avoid revisiting).
    
    Returns:
       A dictionary mapping a path length (an integer) to the number of paths
       (with multiplicative weight) that reach the target with that exact length.
       If no path is found, returns an empty dictionary.
    """    
    if depth > max_depth:
        return {}
    if current == target:
        # We require a nonzero length path.
        return {depth: 1} if depth > 0 else {}

    counts = {}
    for neighbor, mult in get_neighbors(current, H):
        # Skip v_exclude if it is not the target (to ensure v appears only at start/end)
        if neighbor[0] == "v" and neighbor[1] == v_exclude and neighbor != target:
            continue
        # Enforce simplicity: do not revisit nodes (except the target)
        if neighbor != target and neighbor in visited:
            continue
        # Mark neighbor as visited (if not target)
        if neighbor != target:
            visited.add(neighbor)
        sub = dfs_count_paths(neighbor, target, H, v_exclude, visited, depth + 1, max_depth)
        if neighbor != target:
            visited.remove(neighbor)
        # Multiply the counts by the multiplicity of the edge used
        for d, cnt in sub.items():
            counts[d] = counts.get(d, 0) + cnt * mult
    return counts


def shortest_cycle_and_count_for_variable(H, v, max_depth=10):
    """
    For a given parity-check matrix H (possibly with duplicate edges) and a given variable node v,
    compute:
      - lv: the length of the shortest cycle in the Tanner graph that involves v, and
      - mv: the number of cycles (counted with multiplicity) of length lv that involve v.
    
    Here, a cycle is defined as a simple cycle in the bipartite Tanner graph.
    (Cycles of length 2 arise when a check node is incident on v more than once.)
    
    For cycles of length >= 4, we proceed as follows:
      For each edge (v, c), we remove one copy of that edge and count the number
      of simple paths from ('c', c) back to ('v', v) using a DFS (with depth limit max_depth).
      Each such path of length d gives a cycle of length d + 1 (restoring the removed edge).
    
    Since any such cycle touches v by exactly two edges, each cycle is found twice overall.
    
    Returns:
       (lv, mv), where lv is the shortest cycle length (or None if no cycle is found)
       and mv is the number of cycles of that length.
    """
    m = H.shape[0]
    check_neighbors = [c for c in range(m) if H[c, v] > 0]

    best = float("inf")
    ways_sum = 0

    # 2-cycles (multi-edge)
    for c in check_neighbors:
        if H[c, v] > 1:
            # A 2-cycle exists from v -> c -> v.
            if 2 < best:
                best = 2
                ways_sum = 0
            if 2 == best:
                # There are choose(H[c,v], 2) cycles contributed by check node c.
                ways_sum += (H[c, v] * (H[c, v] - 1)) // 2
    if best == 2:
        return 2, ways_sum

    # cycles >= 4
    for c in check_neighbors:
        H[c, v] -= 1    # Remove one copy of edge (v, c)
        # Start DFS from node ('c', c) with target ('v', v).
        # We begin with visited containing the start.
        visited = {("c", c)}
        result = dfs_count_paths(("c", c), ("v", v), H, v_exclude=v, visited=visited, depth=0, max_depth=max_depth)
        H[c, v] += 1    # Restore the edge.
        if result:
            dmin = min(result.keys())
            cycle_len = dmin + 1    # add back the removed edge
            if cycle_len < best:
                best = cycle_len
                ways_sum = result[dmin]
            elif cycle_len == best:
                ways_sum += result[dmin]

    if best == float("inf"):
        return None, 0
    # Each cycle of length >= 4 is counted twice (once for each edge incident on v).
    return best, ways_sum // 2


def _score_key(score):
    """
    Given a score (l, m), return a key for lexicographic comparison.
    We want (l1, m1) < (l2, m2) if l1 < l2 or (l1 == l2 and m1 > m2).
    Mapping (l, m) -> (l, -m) achieves this.
    """    
    l, m = score
    return (l, -m)


def _is_better(new1, new2, old1, old2):
    """
    Compare two pairs of scores for variable nodes v1 and v2.
    Each score is a tuple (l, m).
    
    We first compute:
      new_min = min(new_score_v1, new_score_v2)  (using our custom ordering)
      old_min = min(old_score_v1, old_score_v2)
    
    If new_min is better (i.e. smaller in our order) than old_min, we return True.
    Otherwise, if they are equal, we compare the corresponding maximums and return True
    if the new maximum is better than the old maximum.
    
    Otherwise, return False.
    """    
    new_min = min(new1, new2, key=_score_key)
    old_min = min(old1, old2, key=_score_key)
    if _score_key(new_min) > _score_key(old_min):
        return True
    if _score_key(new_min) == _score_key(old_min):
        new_max = max(new1, new2, key=_score_key)
        old_max = max(old1, old2, key=_score_key)
        return _score_key(new_max) > _score_key(old_max)
    return False


def enumerate_edges(H):
    """
    Return a list of edge instances from the Tanner graph.
    Each edge is represented as a tuple (v, c) where v is the variable node (column index)
    and c is the check node (row index). Duplicate edges appear as multiple entries.
    """    
    m, n = H.shape
    edges = []
    for c in range(m):
        for v in range(n):
            for _ in range(H[c, v]):     # if H[c,v] > 0, add that many copies
                edges.append((v, c))
    return edges


def optimize_ldpc(H: np.ndarray, rounds: int, max_depth: int = 10, verbose: bool = False) -> np.ndarray:
    """
    Given a parity-check matrix H (which may contain duplicate edges),
    perform a number of rounds of random edge shuffles.
    
    For each round:
      - Randomly select two edge instances (v1, c1) and (v2, c2) from the Tanner graph.
      - Compute the current scores (lv, mv) for variable nodes v1 and v2.
      - Replace the edges (v1, c1) and (v2, c2) by (v1, c2) and (v2, c1):
            * Decrease H[c1, v1] and H[c2, v2] by 1.
            * Increase H[c1, v2] and H[c2, v1] by 1.
      - Recompute the scores for v1 and v2.
      - If the new pair of scores is "better" (as determined by is_better), keep the change;
        otherwise, revert the swap.
    
    Returns the modified matrix H.
    """
    H = np.asarray(H, dtype=int, order="C")
    for _ in range(rounds):
        # Enumerate all edge instances.
        edges = enumerate_edges(H)
        if len(edges) < 2:
            break   # Not enough edges to swap.

        # Randomly pick two distinct edge instances.
        (v1, c1), (v2, c2) = random.sample(edges, 2)

        # Save the old scores for v1 and v2.
        old1 = shortest_cycle_and_count_for_variable(H, v1, max_depth)
        old2 = shortest_cycle_and_count_for_variable(H, v2, max_depth)

        # Perform the swap: remove one copy of (v1, c1) and (v2, c2)
        # and add one copy of (v1, c2) and (v2, c1).
        H[c1, v1] -= 1
        H[c2, v2] -= 1
        H[c1, v2] += 1
        H[c2, v1] += 1

        # Recompute the scores for v1 and v2 after the swap.
        new1 = shortest_cycle_and_count_for_variable(H, v1, max_depth)
        new2 = shortest_cycle_and_count_for_variable(H, v2, max_depth)

        # Check the condition: if the new scores are "better" keep the swap, else revert.
        if _is_better(new1, new2, old1, old2):
            # Keep the swap (optionally, print or log the improvement).
            if verbose:
                print(f"{old1},{old2} -> {new1},{new2}")
        else:
            # Revert the swap.
            H[c1, v1] += 1
            H[c2, v2] += 1
            H[c1, v2] -= 1
            H[c2, v1] -= 1
    return H


# ============================================================
# High-level helper: generate an LDPC with "good girth"
# ============================================================

def generate_ldpc_good_girth(
    n: int,
    dv: int,
    dc: int,
    target_girth: int,
    max_outer_iters: int = 10,
    rounds_per_iter: int = 100,
    max_depth: int = 10,
    require_full_row_rank: bool = True,
    require_no_multi_edges: bool = True,
    seed: Optional[int] = None,
) -> Tuple[np.ndarray, int]:
    """
    Generate & optimize an LDPC parity-check matrix until girth >= target_girth
    and (optionally) other constraints are satisfied.

    Returns: (H, girth)
    """
    if seed is not None:
        random.seed(seed)
        np.random.seed(seed)

    H = generate_ldpc(n, dv, dc)

    for iter in range(max_outer_iters):
        H = optimize_ldpc(H, rounds=rounds_per_iter, max_depth=max_depth, verbose=True)
        g = compute_girth_from_parity_check(H)
        print('Iteration {}/{}: Girth = {}'.format(iter + 1, max_outer_iters, g))

        if require_no_multi_edges and has_duplicate_edges(H):
            continue
        if require_full_row_rank and gf2_rank(H % 2) < H.shape[0]:
            continue
        if g >= target_girth:
            return H, g

    # return best-effort
    return H % 2, compute_girth_from_parity_check(H)



from .base import CircuitBuilder


class FreeformBuilder(CircuitBuilder):
    name = "freeform"

    def build(self, code, **opts):
        raise NotImplementedError("Freeform circuit construction is not implemented yet.")



from .base import CircuitBuilder


class XZColorationBuilder(CircuitBuilder):
    name = "xzcoloration"

    def build(self, code, **opts):
        raise NotImplementedError("XZ coloration circuit construction is not implemented yet.")



"""
Lift-connected surface code (LCS) construction.

We parameterize the code by:
- length = l + 1 (number of columns in the base matrix)
- lift_size = L (circulant lift size)

The base matrix has shape (l, l + 1) and uses polynomial entries:
    I  and  I + P
on the two diagonals.
"""

import numpy as np

from .qlp import QlpPolyCode


class LscCode(QlpPolyCode):
    def __init__(self, lift_size, length):
        """
        :param lift_size: L, circulant lift size.
        :param length: l + 1, number of columns in the base matrix.
        """
        # Reference: J. Old, M. Rispler, and M. Muller, arXiv:2401.02911 (lift-connected surface codes).
        if length < 2:
            raise ValueError("length must be at least 2 so that l = length - 1 is positive.")

        l = length - 1
        b = [[[] for _ in range(length)] for _ in range(l)]
        for i in range(l):
            b[i][i] = [0]
            b[i][i + 1] = [0, 1]

        self.length = length
        self.l = l
        self.lift_size = lift_size
        self.b = np.array(b, dtype=object)

        super().__init__(b, b, lift_size)


__all__ = ["LscCode"]



from .cardinal import CardinalBuilder
from .freeform import FreeformBuilder
from .xzcoloration import XZColorationBuilder

_BUILDERS = {
    CardinalBuilder.name: CardinalBuilder,
    XZColorationBuilder.name: XZColorationBuilder,
    FreeformBuilder.name: FreeformBuilder,
}


def get_builder(name):
    if name is None:
        name = "cardinal"
    builder_cls = _BUILDERS.get(name)
    if builder_cls is None:
        raise ValueError(f"Unknown circuit construction strategy: {name}")
    return builder_cls()


__all__ = ["get_builder", "CardinalBuilder", "XZColorationBuilder", "FreeformBuilder"]



import numpy as np
import stim

from quits.circuit import check_overlapping_CX, get_qldpc_mem_circuit
from quits.qldpc_code import HgpCode


def test_check_overlapping_cx_hgp_prints_when_verbose():
    h = np.loadtxt(
        "parity_check_matrices/n=12_dv=3_dc=4_dist=6.txt",
        dtype=int,
    )
    code = HgpCode(h, h)
    code.build_circuit(strategy="cardinal", seed=22)

    circuit = stim.Circuit(
        get_qldpc_mem_circuit(code, 1e-3, 1e-3, 1e-3, 1e-3, 1, basis="Z")
    )

    overlaps = check_overlapping_CX(circuit, verbose=True)
    print({"overlaps": [(i, d.tolist()) for i, d in overlaps]})

    assert overlaps == []



"""
@author: Mingyu Kang, Yingjia Lin
"""

import warnings

import numpy as np
from scipy.linalg import circulant
from ..gf2_util import verify_css_logicals
from .circuit_construction import get_builder


class QldpcCode:
    def __init__(self):

        self.hz, self.hx = None, None
        self.lz, self.lx = None, None

        self.data_qubits, self.zcheck_qubits, self.xcheck_qubits = None, None, None
        self.check_qubits, self.all_qubits = None, None

    def verify_css_logicals(self):
        return verify_css_logicals(self.hz, self.hx, self.lz, self.lx)

    def get_circulant_mat(self, size, power):
        return circulant(np.eye(size, dtype=int)[:, power])

    def lift(self, lift_size, h_base, h_base_placeholder):
        '''
        :param lift_size: Size of cyclic matrix to which each monomial entry is lifted.
        :param h_base: Base matrix where each entry is the power of the monomial.
        :param h_base_placeholder: Placeholder matrix where each non-zero entry of the base matrix is replaced by 1.
        :return: Lifted matrix.
        '''
        h = np.zeros((h_base.shape[0] * lift_size, h_base.shape[1] * lift_size), dtype=int)
        for i in range(h_base.shape[0]):
            for j in range(h_base.shape[1]):
                if h_base_placeholder[i, j] != 0:
                    h[i * lift_size:(i + 1) * lift_size, j * lift_size:(j + 1) * lift_size] = self.get_circulant_mat(
                        lift_size, h_base[i, j]
                    )
        return h

    def lift_enc(self, lift_size, h_base_enc, h_base_placeholder):
        '''
        :param lift_size: Size of cyclic matrix to which each polynomial term is lifted.
        :param h_base: Base matrix where each entry ENCODEs the powers of polynomial terms in base of lift_size.
        :param h_base_placeholder: Placeholder matrix where each non-zero entry of the base matrix is replaced by 1.
        :return: Lifted matrix.
        '''
        h = np.zeros((h_base_enc.shape[0] * lift_size, h_base_enc.shape[1] * lift_size), dtype=int)
        for i in range(h_base_enc.shape[0]):
            for j in range(h_base_enc.shape[1]):
                if h_base_placeholder[i, j] != 0:
                    hij_enc = h_base_enc[i, j]
                    if hij_enc == 0:
                        h[i * lift_size:(i + 1) * lift_size, j * lift_size:(j + 1) * lift_size] = self.get_circulant_mat(
                            lift_size, 0
                        )
                    else:
                        while hij_enc > 0:
                            power = hij_enc % lift_size
                            h[i * lift_size:(i + 1) * lift_size, j * lift_size:(j + 1) * lift_size] += self.get_circulant_mat(
                                lift_size, power
                            )
                            hij_enc = hij_enc // lift_size
        return h

    # Draw the Tanner graph of the code.
    def draw_graph(self, draw_edges=True):
        builder = get_builder("cardinal")
        return builder.draw_graph(self, draw_edges=draw_edges)

    def build_circuit(self, strategy="cardinal", **opts):
        builder = get_builder(strategy)
        return builder.build(self, **opts)

    def build_graph(self, **opts):
        warnings.warn(
            "QldpcCode.build_graph is deprecated; use build_circuit(strategy='cardinal', ...) instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        return self.build_circuit(strategy="cardinal", **opts)

    # Helper function for assigning bool to each edge of the classical code's parity check matrix
    def get_classical_edge_bools(self, h, seed):
        builder = get_builder("cardinal")
        return builder.get_classical_edge_bools(h, seed)

    # Helper function for adding edges
    def add_edge(self, edge_no, direction_ind, control, target):
        builder = get_builder("cardinal")
        return builder.add_edge(self, edge_no, direction_ind, control, target)

    def color_edges(self):
        builder = get_builder("cardinal")
        return builder.color_edges(self)


__all__ = ["QldpcCode"]



"""
@author: Mingyu Kang, Yingjia Lin
"""

import numpy as np

from .circuit_construction import get_builder
from ..gf2_util import _gf2_inv_square, compute_lz_and_lx
from .base import QldpcCode


class BpcCode(QldpcCode):
    def __init__(self, p1, p2, lift_size, factor, verbose=False, canonical_basis="z"):
        '''
        :param p1: First polynomial used to construct the bp code. Each entry of the list is the power of each polynomial term.
                   e.g. p1 = [0, 1, 5] represents the polynomial 1 + x + x^5
        :param p2: Second polynomial used to construct the bp code. Each entry of the list is the power of each polynomial term.
        :param lift_size: Size of cyclic matrix to which each monomial entry is lifted.
        :param factor: Power of the monomial generator of the cyclic subgroup that is factored out by the balanced product.
                       e.g. if factor == 3, cyclic subgroup <x^3> is factored out.
        '''
        # Reference: R. Tiew & N. P. Breuckmann, arXiv:2411.03302 (balanced product cyclic codes).
        # Note: To match the paper, p2 should use lift_size minus the powers listed there (transpose convention).
        super().__init__()

        self.p1, self.p2 = p1, p2
        self.lift_size = lift_size
        self.factor = factor
        self.verbose = verbose
        self.canonical_basis = canonical_basis

        b1 = np.zeros((self.factor, self.factor), dtype=int)
        b1_placeholder = np.zeros((self.factor, self.factor), dtype=int)
        for power in p1:
            mat, mat_placeholder = self.get_block_mat(power)
            b1 = b1 + mat
            b1_placeholder = b1_placeholder + mat_placeholder
        b1T = (self.lift_size - b1.T) % self.lift_size
        b1T_placeholder = b1_placeholder.T

        self.b1, self.b1T = b1, b1T
        self.b1_placeholder, self.b1T_placeholder = b1_placeholder, b1T_placeholder

        h1 = self.lift(self.lift_size, b1, b1_placeholder)
        h1T = self.lift(self.lift_size, b1T, b1T_placeholder)

        h2 = np.zeros((self.lift_size, self.lift_size), dtype=int)
        for power in p2:
            h2 = h2 + self.get_circulant_mat(self.lift_size, power)
        h2 = np.kron(np.eye(self.factor, dtype=int), h2)
        h2T = h2.T

        self.hz = np.concatenate((h2, h1T), axis=1)
        self.hx = np.concatenate((h1, h2T), axis=1)
        # q = lift_size / factor in the balanced product construction.
        q = self.lift_size // self.factor
        if q % 2 == 1:
            if self.verbose:
                print("BpcCode: using canonical logical codewords (q is odd).")
            self.lz, self.lx = self.get_canonical_logicals(canonical_basis=self.canonical_basis)
        else:
            self.lz, self.lx = compute_lz_and_lx(self.hz, self.hx)

    def get_block_mat(self, power):
        gen_mat = self.get_circulant_mat(self.factor, 1)
        gen_mat[0, -1] = 2

        mat = np.linalg.matrix_power(gen_mat, power)
        mat_placeholder = (mat > 0) * 1

        mat = np.log2(mat + 1e-8).astype(int)
        mat = mat * mat_placeholder * self.factor
        return mat, mat_placeholder

    # WRONG; SHOULD BE FIXED LATER
    def get_canonical_logicals(self, canonical_basis="z"):
        '''
        :return: Logical operators of the code as a list of tuples (logical_z, logical_x)
                 where logical_z and logical_x are numpy arrays of shape (num_logicals, num_data_qubits)
                 The logicals are written in the "canonical form" as described in Eq. 30 of arXiv:2411.03302
        '''

        lz = np.zeros((2 * (self.factor - 1) ** 2, self.hz.shape[1]), dtype=int)
        lx = np.zeros((2 * (self.factor - 1) ** 2, self.hx.shape[1]), dtype=int)

        cnt = 0
        for i in range(self.factor - 1):
            for j in range(self.factor - 1):
                yi_vec = self.get_circulant_mat(self.factor, 0)[:, i]
                xjgx_vec = (self.get_circulant_mat(self.factor, 0) + self.get_circulant_mat(self.factor, 1))[:, j]
                xjgx_vec = np.tile(xjgx_vec, self.lift_size // self.factor)

                prod = np.kron(yi_vec, xjgx_vec)
                lz[cnt, :] = np.concatenate((np.zeros(self.hz.shape[1] - len(prod), dtype=int), prod))
                lx[cnt, :] = np.concatenate((prod, np.zeros(self.hx.shape[1] - len(prod), dtype=int)))

                cnt += 1

        for i in range(self.factor - 1):
            for j in range(self.factor - 1):
                yigy_vec = (self.get_circulant_mat(self.factor, 0) + self.get_circulant_mat(self.factor, 1))[:, i]
                xj_vec = self.get_circulant_mat(self.factor, 0)[:, j]
                xj_vec = np.tile(xj_vec, self.lift_size // self.factor)

                prod = np.kron(yigy_vec, xj_vec)
                lz[cnt, :] = np.concatenate((prod, np.zeros(self.hz.shape[1] - len(prod), dtype=int)))
                lx[cnt, :] = np.concatenate((np.zeros(self.hx.shape[1] - len(prod), dtype=int), prod))

                cnt += 1

        if canonical_basis == "z":
            pairing = (lz @ lx.T) & 1
            inv_pairing = _gf2_inv_square(pairing)
            lx = (inv_pairing.T @ lx) & 1
        elif canonical_basis == "x":
            pairing = (lx @ lz.T) & 1
            inv_pairing = _gf2_inv_square(pairing)
            lz = (inv_pairing @ lz) & 1

        return lz, lx

    def build_circuit(self, strategy="cardinal", **opts):
        if strategy != "cardinal":
            return super().build_circuit(strategy=strategy, **opts)
        return self._build_cardinal_graph(**opts)

    def _build_cardinal_graph(self, seed=1):
        get_builder("cardinal").build_graph(self)
        data_qubits, zcheck_qubits, xcheck_qubits = [], [], []

        # Add nodes to the Tanner graph
        for i in range(self.factor):
            for l in range(self.lift_size):
                node = i * self.lift_size + l
                data_qubits += [node]
                self.graph.add_node(node, pos=(2 * i, 0))
                self.node_colors += ['blue']

        start = self.factor * self.lift_size
        for i in range(self.factor):
            for l in range(self.lift_size):
                node = start + i * self.lift_size + l
                xcheck_qubits += [node]
                self.graph.add_node(node, pos=(2 * i + 1, 0))
                self.node_colors += ['purple']

        start = 2 * self.factor * self.lift_size
        for i in range(self.factor):
            for l in range(self.lift_size):
                node = start + i * self.lift_size + l
                zcheck_qubits += [node]
                self.graph.add_node(node, pos=(2 * i, 1))
                self.node_colors += ['green']

        start = 3 * self.factor * self.lift_size
        for i in range(self.factor):
            for l in range(self.lift_size):
                node = start + i * self.lift_size + l
                data_qubits += [node]
                self.graph.add_node(node, pos=(2 * i + 1, 1))
                self.node_colors += ['blue']

        self.data_qubits = sorted(np.array(data_qubits))
        self.zcheck_qubits = sorted(np.array(zcheck_qubits))
        self.xcheck_qubits = sorted(np.array(xcheck_qubits))
        self.check_qubits = np.concatenate((self.zcheck_qubits, self.xcheck_qubits))
        self.all_qubits = sorted(np.array(data_qubits + zcheck_qubits + xcheck_qubits))

        hedge_bool_list = self.get_classical_edge_bools(np.ones(self.b1.shape, dtype=int), seed)
        vedge_bool_list = self.get_classical_edge_bools(np.ones(self.b1.shape, dtype=int), seed)

        # Add edges to the Tanner graph of each direction
        edge_no = 0
        for i in range(self.factor):
            for j in range(self.factor):
                shift = self.b1[i, j]
                edge_bool = hedge_bool_list[(i, j)]

                for l in range(self.lift_size):
                    for k in range(2):  # 0 : bottom, 1 : top
                        if k ^ edge_bool:
                            direction_ind = self.direction_inds['E']
                        else:
                            direction_ind = self.direction_inds['W']

                        control = (2 * k + 1) * self.factor * self.lift_size + i * self.lift_size + (l + shift) % self.lift_size
                        target = 2 * k * self.factor * self.lift_size + j * self.lift_size + l
                        self.add_edge(edge_no, direction_ind, control, target)
                        edge_no += 1

        def shuffle(node_no, qubit_no):
            m, r = qubit_no // self.factor, qubit_no % self.factor
            return r, self.lift_size // self.factor * node_no + m

        for i in range(self.factor):
            for j in range(len(self.p2)):
                shift = self.p2[j]

                for l in range(self.lift_size):
                    for k in range(2):  # 0 : left, 1 : right
                        i_shuffled, _ = shuffle(i, l)
                        j_shuffled, _ = shuffle(i, (l + shift) % self.lift_size)
                        edge_bool = vedge_bool_list[(i_shuffled, j_shuffled)]
                        if k ^ edge_bool:
                            direction_ind = self.direction_inds['N']
                        else:
                            direction_ind = self.direction_inds['S']

                        control = k * self.factor * self.lift_size + i * self.lift_size + l
                        target = (2 + k) * self.factor * self.lift_size + i * self.lift_size + (l + shift) % self.lift_size
                        self.add_edge(edge_no, direction_ind, control, target)
                        edge_no += 1

        # Color the edges of self.graph
        self.color_edges()
        return


__all__ = ["BpcCode"]



import numpy as np
import stim
from ldpc.bposd_decoder import BpOsdDecoder

from quits.circuit import get_qldpc_mem_circuit
from quits.decoder.sliding_window import sliding_window_phenom_mem
from quits.qldpc_code import BpcCode, HgpCode, LscCode, QlpCode
from quits.simulation import get_stim_mem_result


def _simulate_mem_circuit(code, p, num_rounds, num_trials, basis="Z"):
    circuit = stim.Circuit(
        get_qldpc_mem_circuit(
            code,
            p,
            p,
            p,
            p,
            num_rounds,
            basis=basis,
        )
    )
    detection_events, observable_flips = get_stim_mem_result(
        circuit,
        num_trials,
        seed=1,
    )
    return circuit, detection_events, observable_flips


def _bp_osd_params(max_iter, osd_order):
    return {
        "bp_method": "product_sum",
        "max_iter": max_iter,
        "schedule": "serial",
        "osd_method": "osd_cs",
        "osd_order": osd_order,
    }


def _run_sliding_window_phenom(code, code_name, p, num_rounds, num_trials, W, F, max_iter, osd_order, seed=1):
    code.build_circuit(strategy="cardinal", seed=seed)
    report = code.verify_css_logicals()
    print(f"{code_name} verify_css_logicals", report)
    assert report["ok"]
    depth = sum(code.num_colors.values())
    eff_error_rate_per_fault = p * (depth + 3)

    _, detection_events, observable_flips = _simulate_mem_circuit(
        code,
        p,
        num_rounds,
        num_trials,
    )

    dict1 = _bp_osd_params(max_iter, osd_order)
    dict2 = _bp_osd_params(max_iter, osd_order)
    dict1["error_rate"] = float(eff_error_rate_per_fault)
    dict2["error_rate"] = float(eff_error_rate_per_fault)

    logical_pred = sliding_window_phenom_mem(
        detection_events,
        code.hz,
        code.lz,
        W,
        F,
        BpOsdDecoder,
        BpOsdDecoder,
        dict1,
        dict2,
        "decode",
        "decode",
        tqdm_on=False,
    )

    pL = np.mean((observable_flips - logical_pred).any(axis=1))
    lfr = 1 - (1 - pL) ** (1 / num_rounds)
    return depth, eff_error_rate_per_fault, pL, lfr


def _print_results(code_name, params, depth, eff_error_rate_per_fault, pL, lfr):
    print(
        f"{code_name} sliding_window_phenom_mem",
        {
            **params,
            "depth": depth,
            "eff_error_rate_per_fault": float(eff_error_rate_per_fault),
            "pL": float(pL),
            "lfr": float(lfr),
        },
    )


def test_hgp_code_circuit_phenom_low_lfr():
    h = np.loadtxt(
        "parity_check_matrices/n=12_dv=3_dc=4_dist=6.txt",
        dtype=int,
    )
    code = HgpCode(h, h)

    params = {
        "p": 1e-3,
        "num_rounds": 15,
        "num_trials": 50,
        "W": 5,
        "F": 3,
        "max_iter": 10,
        "osd_order": 1,
    }

    depth, eff_error_rate_per_fault, pL, lfr = _run_sliding_window_phenom(code, "HGP", **params, seed=22)
    _print_results("HGP", params, depth, eff_error_rate_per_fault, pL, lfr)
    print()

    assert pL <= 0.25
    assert lfr <= 0.08


def test_qlp_code_circuit_phenom_low_lfr():
    lift_size = 16
    b = np.array(
        [
            [0, 0, 0, 0, 0],
            [0, 2, 4, 7, 11],
            [0, 3, 10, 14, 15],
        ]
    )
    code = QlpCode(b, b, lift_size)

    params = {
        "p": 5e-4,
        "num_rounds": 15,
        "num_trials": 50,
        "W": 5,
        "F": 3,
        "max_iter": 10,
        "osd_order": 1,
    }

    depth, eff_error_rate_per_fault, pL, lfr = _run_sliding_window_phenom(code, "QLP", **params)
    _print_results("QLP", params, depth, eff_error_rate_per_fault, pL, lfr)
    print()

    assert pL <= 0.3
    assert lfr <= 0.1


def test_bpc_code_circuit_phenom_low_lfr():
    lift_size, factor = 15, 3
    p1 = [0, 1, 5]
    p2 = [0, 8, 13]
    code = BpcCode(p1, p2, lift_size, factor)

    params = {
        "p": 1e-3,
        "num_rounds": 15,
        "num_trials": 50,
        "W": 5,
        "F": 3,
        "max_iter": 10,
        "osd_order": 1,
    }

    depth, eff_error_rate_per_fault, pL, lfr = _run_sliding_window_phenom(code, "BPC", **params)
    _print_results("BPC", params, depth, eff_error_rate_per_fault, pL, lfr)
    print()

    assert pL <= 0.3
    assert lfr <= 0.1


def test_lsc_code_circuit_phenom_low_lfr():
    lift_size = 5
    length = 3  # length = l + 1
    code = LscCode(lift_size, length)

    expected_b = np.array(
        [
            [[0], [0, 1], []],
            [[], [0], [0, 1]],
        ],
        dtype=object,
    )
    assert (code.b == expected_b).all()

    params = {
        "p": 1e-3,
        "num_rounds": 15,
        "num_trials": 50,
        "W": 5,
        "F": 3,
        "max_iter": 10,
        "osd_order": 1,
    }

    depth, eff_error_rate_per_fault, pL, lfr = _run_sliding_window_phenom(code, "LSC", **params, seed=1)
    _print_results("LSC", params, depth, eff_error_rate_per_fault, pL, lfr)
    print()

    assert pL <= 0.3
    assert lfr <= 0.1



import numpy as np
import stim

from quits.circuit import get_qldpc_mem_circuit
from quits.decoder import sliding_window_bplsd_phenom_mem, sliding_window_bposd_phenom_mem
from quits.qldpc_code import HgpCode
from quits.simulation import get_stim_mem_result


def _build_hgp_code(seed=22):
    h = np.loadtxt(
        "parity_check_matrices/n=12_dv=3_dc=4_dist=6.txt",
        dtype=int,
    )
    code = HgpCode(h, h)
    code.build_circuit(strategy="cardinal", seed=seed)
    return code


def _simulate_mem_circuit(code, p, num_rounds, num_trials, basis="Z"):
    circuit = stim.Circuit(
        get_qldpc_mem_circuit(
            code,
            p,
            p,
            p,
            p,
            num_rounds,
            basis=basis,
        )
    )
    detection_events, observable_flips = get_stim_mem_result(
        circuit,
        num_trials,
        seed=1,
    )
    return circuit, detection_events, observable_flips


def _run_phenom_decoder(
    decoder_fn,
    decoder_params,
    code,
    p,
    num_rounds,
    num_trials,
    W,
    F,
):
    depth = sum(code.num_colors.values())
    eff_error_rate_per_fault = p * (depth + 3)

    _, detection_events, observable_flips = _simulate_mem_circuit(
        code,
        p,
        num_rounds,
        num_trials,
    )

    logical_pred = decoder_fn(
        detection_events,
        code.hz,
        code.lz,
        W,
        F,
        eff_error_rate_per_fault=eff_error_rate_per_fault,
        tqdm_on=False,
        **decoder_params,
    )

    pL = np.mean((observable_flips - logical_pred).any(axis=1))
    lfr = 1 - (1 - pL) ** (1 / num_rounds)
    return depth, eff_error_rate_per_fault, pL, lfr


def _print_results(name, params, depth, eff_error_rate_per_fault, pL, lfr):
    print(
        f"{name} sliding_window_phenom_mem",
        {
            **params,
            "depth": depth,
            "eff_error_rate_per_fault": float(eff_error_rate_per_fault),
            "pL": float(pL),
            "lfr": float(lfr),
        },
    )


def test_bposd_decoder_phenom_low_lfr():
    code = _build_hgp_code(seed=22)
    report = code.verify_css_logicals()
    assert report["ok"]

    params = {
        "p": 1e-3,
        "num_rounds": 15,
        "num_trials": 50,
        "W": 5,
        "F": 3,
        "max_iter": 10,
        "osd_order": 1,
    }

    decoder_params = {
        "max_iter": params["max_iter"],
        "osd_order": params["osd_order"],
    }

    depth, eff_error_rate_per_fault, pL, lfr = _run_phenom_decoder(
        sliding_window_bposd_phenom_mem,
        decoder_params,
        code,
        params["p"],
        params["num_rounds"],
        params["num_trials"],
        params["W"],
        params["F"],
    )
    _print_results("BPOSD", params, depth, eff_error_rate_per_fault, pL, lfr)
    print()

    assert pL <= 0.25
    assert lfr <= 0.08


def test_bplsd_decoder_phenom_low_lfr():
    code = _build_hgp_code(seed=22)
    report = code.verify_css_logicals()
    assert report["ok"]

    params = {
        "p": 1e-3,
        "num_rounds": 15,
        "num_trials": 50,
        "W": 5,
        "F": 3,
        "max_iter": 10,
        "lsd_order": 1,
    }

    decoder_params = {
        "max_iter": params["max_iter"],
        "lsd_order": params["lsd_order"],
    }

    depth, eff_error_rate_per_fault, pL, lfr = _run_phenom_decoder(
        sliding_window_bplsd_phenom_mem,
        decoder_params,
        code,
        params["p"],
        params["num_rounds"],
        params["num_trials"],
        params["W"],
        params["F"],
    )
    _print_results("BPLSD", params, depth, eff_error_rate_per_fault, pL, lfr)
    print()

    assert pL <= 0.3
    assert lfr <= 0.1



"""Decoder utilities and sliding window implementations."""

from .base import (
    detector_error_model_to_matrix,
    dict_to_csc_matrix_column_row,
    dict_to_csc_matrix_row_column,
    spacetime,
)
from .bplsd import sliding_window_bplsd_circuit_mem, sliding_window_bplsd_phenom_mem
from .bposd import sliding_window_bposd_circuit_mem, sliding_window_bposd_phenom_mem
from .sliding_window import sliding_window_circuit_mem, sliding_window_phenom_mem

__all__ = [
    "detector_error_model_to_matrix",
    "dict_to_csc_matrix_column_row",
    "dict_to_csc_matrix_row_column",
    "spacetime",
    "sliding_window_phenom_mem",
    "sliding_window_bposd_phenom_mem",
    "sliding_window_bplsd_phenom_mem",
    "sliding_window_circuit_mem",
    "sliding_window_bposd_circuit_mem",
    "sliding_window_bplsd_circuit_mem",
]



"""
@author: Mingyu Kang, Yingjia Lin
"""

import numpy as np

from .circuit_construction import get_builder
from ..gf2_util import gf2_nullspace_basis, gf2_coset_reps_rowspace, compute_lz_and_lx
from .base import QldpcCode


class HgpCode(QldpcCode):
    def __init__(self, h1, h2, verbose=False):
        '''
        :param h1: Parity check matrix of the first classical code used to construct the hgp code
        :param h2: Parity check matrix of the second classical code used to construct the hgp code
        '''
        # Reference: Tillich & Zemor, arXiv:0903.0566 (hypergraph product codes).
        # Canonical logicals follow the HGP convention in arXiv:2204.10812.
        super().__init__()

        self.h1, self.h2 = h1, h2
        self.verbose = verbose
        self.r1, self.n1 = h1.shape
        self.r2, self.n2 = h2.shape

        self.hz = np.concatenate((np.kron(h2, np.eye(self.n1, dtype=int)),
                                  np.kron(np.eye(self.r2, dtype=int), h1.T)), axis=1)
        self.hx = np.concatenate((np.kron(np.eye(self.n2, dtype=int), h1),
                                  np.kron(h2.T, np.eye(self.r1, dtype=int))), axis=1)

        self.l1 = gf2_nullspace_basis(self.h1)
        self.l2 = gf2_nullspace_basis(self.h2)
        self.k1, self.k2 = self.l1.shape[0], self.l2.shape[0]

        if self.r1 == self.n1 - self.k1 and self.r2 == self.n2 - self.k2:  # If both classical parity check matrices are full-rank
            if self.verbose:
                print("HgpCode: using canonical logical codewords.")
            self.lz, self.lx = self.get_canonical_logicals()  # set logical operators in the canonical form
        else:
            self.lz, self.lx = compute_lz_and_lx(self.hz, self.hx)

    def get_canonical_logicals(self):
        """
        Canonical logicals for your HGP convention, assuming k1^T=k2^T=0.
        Returns:
          lz, lx: shape (k1*k2, num_data_qubits) as uint8.
        """
        E1 = gf2_coset_reps_rowspace(self.h1)  # (k1, n1) if H1 full row rank
        E2 = gf2_coset_reps_rowspace(self.h2)  # (k2, n2)

        lz = np.zeros((self.k1 * self.k2, self.hz.shape[1]), dtype=np.uint8)
        lx = np.zeros((self.k1 * self.k2, self.hx.shape[1]), dtype=np.uint8)

        cnt = 0
        for i in range(self.k2):
            for j in range(self.k1):
                # Z: (E2_i ⊗ L1_j | 0)
                lz[cnt, :self.n1 * self.n2] = np.kron(E2[i, :], self.l1[j, :]) & 1
                # X: (L2_i ⊗ E1_j | 0)
                lx[cnt, :self.n1 * self.n2] = np.kron(self.l2[i, :], E1[j, :]) & 1
                cnt += 1

        return lz, lx

    def build_circuit(self, strategy="cardinal", **opts):
        if strategy != "cardinal":
            return super().build_circuit(strategy=strategy, **opts)
        return self._build_cardinal_graph(**opts)

    def _build_cardinal_graph(self, seed=1):
        get_builder("cardinal").build_graph(self)
        data_qubits, zcheck_qubits, xcheck_qubits = [], [], []

        # Add nodes to the Tanner graph
        for i in range(self.n1):
            for j in range(self.n2):
                node = i + j * (self.n1 + self.r1)
                data_qubits += [node]
                self.graph.add_node(node, pos=(i, j))
                self.node_colors += ['blue']

        start = self.n1
        for i in range(self.r1):
            for j in range(self.n2):
                node = start + i + j * (self.n1 + self.r1)
                xcheck_qubits += [node]
                self.graph.add_node(node, pos=(i + self.n1, j))
                self.node_colors += ['purple']

        start = self.n2 * (self.n1 + self.r1)
        for i in range(self.n1):
            for j in range(self.r2):
                node = start + i + j * (self.n1 + self.r1)
                zcheck_qubits += [node]
                self.graph.add_node(node, pos=(i, j + self.n2))
                self.node_colors += ['green']

        start = self.n2 * (self.n1 + self.r1) + self.n1
        for i in range(self.r1):
            for j in range(self.r2):
                node = start + i + j * (self.n1 + self.r1)
                data_qubits += [node]
                self.graph.add_node(node, pos=(i + self.n1, j + self.n2))
                self.node_colors += ['blue']

        self.data_qubits = sorted(np.array(data_qubits))
        self.zcheck_qubits = sorted(np.array(zcheck_qubits))
        self.xcheck_qubits = sorted(np.array(xcheck_qubits))
        self.check_qubits = np.concatenate((self.zcheck_qubits, self.xcheck_qubits))
        self.all_qubits = sorted(np.array(data_qubits + zcheck_qubits + xcheck_qubits))

        hedge_bool_list = self.get_classical_edge_bools(self.h1, seed)
        vedge_bool_list = self.get_classical_edge_bools(self.h2, seed)

        edge_no = 0
        for classical_edge in np.argwhere(self.h1 == 1):
            c0, c1 = classical_edge
            edge_bool = hedge_bool_list[(c0, c1)]
            for k in range(self.n2 + self.r2):
                control, target = (k * (self.n1 + self.r1) + c0 + self.n1, k * (self.n1 + self.r1) + c1)
                if (k < self.n2) ^ edge_bool:
                    direction_ind = self.direction_inds['E']
                else:
                    direction_ind = self.direction_inds['W']
                self.add_edge(edge_no, direction_ind, control, target)
                edge_no += 1

        for classical_edge in np.argwhere(self.h2 == 1):
            c0, c1 = classical_edge
            edge_bool = vedge_bool_list[(c0, c1)]
            for k in range(self.n1 + self.r1):
                control, target = (k + c1 * (self.n1 + self.r1), k + (c0 + self.n2) * (self.n1 + self.r1))
                if (k < self.n1) ^ edge_bool:
                    direction_ind = self.direction_inds['N']
                else:
                    direction_ind = self.direction_inds['S']
                self.add_edge(edge_no, direction_ind, control, target)
                edge_no += 1

        # Color the edges of self.graph
        self.color_edges()
        return


__all__ = ["HgpCode"]



# This code reimplements and adapts several GF(2) linear-algebra and CSS
# logical-operator routines from the QuantumGizmos ldpc package
# (https://github.com/quantumgizmos/ldpc), in particular the file
#   - src_python/ldpc/mod2/mod2_numpy.py


import numpy as np
from typing import Tuple, Dict


# ============================================================
# Minimal GF(2) linear algebra utilities (numpy)
# ============================================================

def _as_gf2(A) -> np.ndarray:
    """Convert input to uint8 array reduced mod 2."""
    return (np.asarray(A) & 1).astype(np.uint8, copy=False)


def gf2_rref(H: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
    """
    Reduced row echelon form over GF(2).

    Returns:
      R: RREF(H) as uint8
      pivots: pivot column indices as int array
    """
    A = _as_gf2(H).copy()
    m, n = A.shape
    pivots = []
    r = 0
    for c in range(n):
        if r >= m:
            break
        rows = np.where(A[r:, c] == 1)[0]
        if rows.size == 0:
            continue
        p = r + int(rows[0])
        if p != r:
            A[[r, p], :] = A[[p, r], :]
        # eliminate pivot column in all other rows (RREF)
        ones = np.where(A[:, c] == 1)[0]
        ones = ones[ones != r]
        if ones.size:
            A[ones, :] ^= A[r, :]
        pivots.append(c)
        r += 1
    return A, np.array(pivots, dtype=int)


def gf2_rank(H: np.ndarray) -> int:
    """Rank over GF(2)."""
    return int(gf2_rref(H)[1].size)


def gf2_row_basis(H: np.ndarray) -> np.ndarray:
    """
    Row basis for rowspace(H) over GF(2).
    Returned as nonzero rows of RREF(H).
    """
    R, _ = gf2_rref(H)
    nz = np.where(np.any(R == 1, axis=1))[0]
    return R[nz, :].astype(np.uint8, copy=False)


def gf2_nullspace_basis(H: np.ndarray) -> np.ndarray:
    """
    Basis for the (right) nullspace of H over GF(2): {x : H x = 0}.
    Returned as rows. Shape: (nullity, n).
    """
    H = _as_gf2(H)
    if H.ndim != 2:
        raise ValueError("H must be 2D")
    _, n = H.shape

    R, pivots = gf2_rref(H)
    pivset = set(pivots.tolist())
    free = [c for c in range(n) if c not in pivset]

    if not free:
        return np.zeros((0, n), dtype=np.uint8)

    rank = pivots.size
    basis = []
    for f in free:
        x = np.zeros(n, dtype=np.uint8)
        x[f] = 1
        # Since R is RREF: for pivot row i with pivot column pivots[i],
        # equation is x[pivot] + sum_{c in free} R[i,c] x[c] = 0
        # => x[pivot] = sum_{c in free} R[i,c] x[c]
        for i in range(rank):
            pc = pivots[i]
            s = np.uint8(0)
            for c in free:
                if R[i, c]:
                    s ^= x[c]
            x[pc] = s
        basis.append(x)

    return np.stack(basis, axis=0)


def gf2_coset_reps_rowspace(H: np.ndarray) -> np.ndarray:
    """
    Canonical reps for F2^n / rowspace(H):
    pick standard basis e_j for non-pivot columns of RREF(H).
    """
    H = _as_gf2(H)
    n = H.shape[1]
    piv = set(gf2_rref(H)[1].tolist())
    nonpiv = [c for c in range(n) if c not in piv]
    E = np.zeros((len(nonpiv), n), dtype=np.uint8)
    for t, c in enumerate(nonpiv):
        E[t, c] = 1
    return E


def gf2_row_span(V: np.ndarray) -> np.ndarray:
    """
    Enumerate all NONZERO linear combinations of the rows of V over GF(2).
    WARNING: exponential in k = number of rows. Returns 2^k - 1 vectors.
    """
    V = _as_gf2(V)
    k, n = V.shape
    if k == 0:
        return np.zeros((0, n), dtype=np.uint8)

    out = np.zeros((2**k - 1, n), dtype=np.uint8)

    # Gray-code accumulation: successive masks differ by one bit -> one XOR update
    acc = np.zeros(n, dtype=np.uint8)
    prev_gray = 0
    idx = 0
    for mask in range(1, 2**k):
        gray = mask ^ (mask >> 1)
        diff = gray ^ prev_gray
        bit = (diff & -diff).bit_length() - 1
        acc ^= V[bit]
        out[idx] = acc
        idx += 1
        prev_gray = gray

    return out


def gf2_solve(A: np.ndarray, b: np.ndarray):
    """
    Solve A x = b over GF(2). Return one solution x or None if infeasible.

    A: (m,n), b: (m,)
    """
    A = _as_gf2(A).copy()
    b = _as_gf2(b).reshape(-1).copy()

    m, n = A.shape
    if b.shape[0] != m:
        raise ValueError("dimension mismatch: b must have length m")

    Aug = np.concatenate([A, b[:, None]], axis=1)

    pivot_cols = []
    r = 0
    for c in range(n):
        if r >= m:
            break
        rows = np.where(Aug[r:, c] == 1)[0]
        if rows.size == 0:
            continue
        p = r + int(rows[0])
        if p != r:
            Aug[[r, p], :] = Aug[[p, r], :]
        ones = np.where(Aug[:, c] == 1)[0]
        ones = ones[ones != r]
        if ones.size:
            Aug[ones, c:] ^= Aug[r, c:]
        pivot_cols.append(c)
        r += 1

    # infeasible row: 0...0 | 1
    if np.any(np.all(Aug[:, :n] == 0, axis=1) & (Aug[:, n] == 1)):
        return None

    x = np.zeros(n, dtype=np.uint8)
    for rr, pc in enumerate(pivot_cols):
        x[pc] = Aug[rr, n]
    return x


def in_rowspace(v: np.ndarray, H: np.ndarray) -> bool:
    """
    Check if v is in rowspace(H) over GF(2) by solving H^T a = v.
    """
    v = _as_gf2(v).reshape(-1)
    H = _as_gf2(H)
    return gf2_solve(H.T, v) is not None


# ============================================================
# CSS logicals + classical code's distance 
# ============================================================

def compute_lz(hz: np.ndarray, hx: np.ndarray) -> np.ndarray:
    """
    Compute logical Z operators for CSS code defined by hx, hz:

      ker(hx) \\ rowspace(hz)

    Implemented by stacking [row_basis(hz); nullspace(hx)] and selecting the
    kernel rows whose *row indices* are pivot columns of RREF(stack.T).
    """
    ker_hx = gf2_nullspace_basis(hx)   # rows: basis for ker(hx)
    im_hz  = gf2_row_basis(hz)         # rows: basis for rowspace(hz)

    log_stack = np.vstack([im_hz, ker_hx]).astype(np.uint8, copy=False)

    # Pivots of stack.T: columns correspond to rows of stack
    _, piv = gf2_rref(log_stack.T)
    piv = set(piv.tolist())

    image_rank = im_hz.shape[0]
    keep = [i for i in range(image_rank, log_stack.shape[0]) if i in piv]
    return log_stack[keep, :]


def _gf2_inv_square(A: np.ndarray) -> np.ndarray:
    """Inverse of a full-rank square matrix over GF(2) via Gauss-Jordan."""
    A = (np.asarray(A) & 1).astype(np.uint8, copy=True)
    n, m = A.shape
    if n != m:
        raise ValueError("A must be square")
    Aug = np.concatenate([A, np.eye(n, dtype=np.uint8)], axis=1)

    r = 0
    for c in range(n):
        rows = np.where(Aug[r:, c] == 1)[0]
        if rows.size == 0:
            continue
        p = r + int(rows[0])
        if p != r:
            Aug[[r, p], :] = Aug[[p, r], :]
        ones = np.where(Aug[:, c] == 1)[0]
        ones = ones[ones != r]
        if ones.size:
            Aug[ones, :] ^= Aug[r, :]
        r += 1
        if r == n:
            break

    if not np.array_equal(Aug[:, :n], np.eye(n, dtype=np.uint8)):
        raise RuntimeError("Matrix not invertible over GF(2).")
    return Aug[:, n:]


def compute_lz_and_lx(hz: np.ndarray, hx: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
    """
    Return (lz, lx) where:
      - lz is computed by compute_lz
      - lx is then *dualized* so that (lz @ lx.T) % 2 == I_k,
        with lx chosen from ker(hz).

    This guarantees correct pairing whenever (hx,hz) define a valid CSS code.
    """
    hz = _as_gf2(hz)
    hx = _as_gf2(hx)

    if hz.ndim != 2 or hx.ndim != 2:
        raise ValueError("hz and hx must be 2D arrays")
    if hz.shape[1] != hx.shape[1]:
        raise ValueError("hz and hx must have the same number of columns")
    if np.any((hx @ hz.T) & 1):
        raise ValueError("Not a CSS pair: hx @ hz.T != 0 (mod 2)")

    n = hx.shape[1]
    k = n - gf2_rank(hx) - gf2_rank(hz)
    if k < 0:
        raise ValueError(f"Computed k={k}<0. Checks inconsistent?")
    if k == 0:
        return np.zeros((0, n), dtype=np.uint8), np.zeros((0, n), dtype=np.uint8)

    # 1) Obtain LZ
    lz = compute_lz(hz, hx)
    if lz.shape[0] != k:
        raise RuntimeError(f"compute_lz returned {lz.shape[0]} logical Zs, expected k={k}")

    # 2) Build LX ⊂ ker(hz) such that LX LZ^T = I
    ker_hz = gf2_nullspace_basis(hz)  # rows span ker(hz), dim = n - rank(hz)
    W = (ker_hz @ lz.T) & 1           # (dim ker_hz, k)

    if gf2_rank(W) < k:
        raise RuntimeError(
            "Cannot dualize: ker(hz) does not contain enough vectors with "
            "independent commutation against the chosen lz."
        )

    # Pick k independent rows of W (and corresponding kernel vectors)
    chosen = []
    piv = {}  # pivot col -> vector (as in gf2_row_basis style)
    for i in range(W.shape[0]):
        v = W[i].copy()
        while True:
            ones = np.flatnonzero(v)
            if ones.size == 0:
                break
            c = int(ones[0])
            if c in piv:
                v ^= piv[c]
            else:
                piv[c] = v
                chosen.append(i)
                break
        if len(chosen) == k:
            break

    chosen = np.array(chosen, dtype=int)
    W_sel = W[chosen, :]          # (k,k) full rank
    K_sel = ker_hz[chosen, :]     # (k,n)

    # Want LX so that (LX @ LZ^T)=I.
    # If we set LX = (W_sel^{-1}) @ K_sel then:
    #   LX LZ^T = (W_sel^{-1} K_sel) LZ^T = W_sel^{-1} (K_sel LZ^T) = W_sel^{-1} W_sel = I
    W_inv = _gf2_inv_square(W_sel)
    lx = (W_inv @ K_sel) & 1

    # sanity
    if not np.array_equal((lz @ lx.T) & 1, np.eye(k, dtype=np.uint8)):
        raise RuntimeError("Internal error: dualization failed to produce lz @ lx.T = I.")

    return lz.astype(np.uint8), lx.astype(np.uint8)


def compute_code_distance(H: np.ndarray):
    """
    Compute code distance from parity check matrix H over GF(2):
    """
    ker = gf2_nullspace_basis(H)
    if ker.shape[0] == 0:
        return np.inf
    cw = gf2_row_span(ker)
    return int(np.min(np.sum(cw, axis=1)))

def verify_css_logicals(
    hz: np.ndarray,
    hx: np.ndarray,
    lz: np.ndarray,
    lx: np.ndarray,
) -> Dict[str, object]:
    """
    Verify that user-supplied CSS logical operators (lz, lx) are valid and complete
    for the CSS stabilizers (hz, hx), in the same spirit as verify_fixed_logical_Xs.

    Assumes gf2_util.py provides:
      - _as_gf2
      - gf2_rank
      - gf2_nullspace_basis

    Returns a dict with booleans + ranks + dims + final 'ok'.
    """
    hz = _as_gf2(hz)
    hx = _as_gf2(hx)
    lz = _as_gf2(lz)
    lx = _as_gf2(lx)

    n = hz.shape[1]
    if hx.shape[1] != n or lx.shape[1] != n or lz.shape[1] != n:
        raise ValueError("hz, hx, lz, lx must all have the same number of columns n")

    report: Dict[str, object] = {}

    # --- CSS commutation ---
    css_ok = not np.any((hx @ hz.T) & 1)
    report["css_condition"] = css_ok

    # --- Logical commutation with opposite stabilizers ---
    lz_commutes_with_X = not np.any((hx @ lz.T) & 1)
    lx_commutes_with_Z = not np.any((hz @ lx.T) & 1)
    report["lz_commutes_with_X"] = lz_commutes_with_X
    report["lx_commutes_with_Z"] = lx_commutes_with_Z

    # --- Ranks / expected k ---
    rank_hz = gf2_rank(hz)
    rank_hx = gf2_rank(hx)
    rank_lz = gf2_rank(lz)
    rank_lx = gf2_rank(lx)

    report["rank_hz"] = int(rank_hz)
    report["rank_hx"] = int(rank_hx)
    report["rank_lz"] = int(rank_lz)
    report["rank_lx"] = int(rank_lx)

    k_expected = int(n - rank_hx - rank_hz)
    report["k_expected"] = k_expected

    # --- Independence modulo stabilizers ---
    rank_hz_lz = gf2_rank(np.vstack([hz, lz]))
    rank_hx_lx = gf2_rank(np.vstack([hx, lx]))
    lz_indep_mod_stab = (rank_hz_lz == rank_hz + rank_lz)
    lx_indep_mod_stab = (rank_hx_lx == rank_hx + rank_lx)

    report["lz_independent_mod_Z_stabilizers"] = lz_indep_mod_stab
    report["lx_independent_mod_X_stabilizers"] = lx_indep_mod_stab
    report["rank_hz_plus_lz"] = int(rank_hz_lz)
    report["rank_hx_plus_lx"] = int(rank_hx_lx)

    # --- Spanning normalizers (ker spaces) ---
    ker_hz = gf2_nullspace_basis(hz)
    ker_hx = gf2_nullspace_basis(hx)
    dim_ker_hz = int(ker_hz.shape[0])  # = n - rank(hz)
    dim_ker_hx = int(ker_hx.shape[0])  # = n - rank(hx)

    report["dim_ker_hz"] = dim_ker_hz
    report["dim_ker_hx"] = dim_ker_hx

    spans_ker_hz = (rank_hx_lx == dim_ker_hz)
    spans_ker_hx = (rank_hz_lz == dim_ker_hx)
    report["hx_plus_lx_spans_ker_hz"] = spans_ker_hz
    report["hz_plus_lz_spans_ker_hx"] = spans_ker_hx

    # --- Pairing matrix ---
    overlap_weight = lx @ lz.T
    pairing = overlap_weight % 2
    pairing_rank = int(gf2_rank(pairing)) if (lx.shape[0] and lz.shape[0]) else 0
    report["pairing_rank"] = pairing_rank
    pairing_is_identity = (
        pairing.shape[0] == pairing.shape[1]
        and np.array_equal(pairing, np.eye(pairing.shape[0], dtype=np.uint8))
    )
    report["pairing_is_identity"] = pairing_is_identity

    # --- Final verdict ---
    ok = (
        css_ok
        and lz_commutes_with_X
        and lx_commutes_with_Z
        and (rank_lz == k_expected)
        and (rank_lx == k_expected)
        and lz_indep_mod_stab
        and lx_indep_mod_stab
        and spans_ker_hz
        and spans_ker_hx
        and (pairing_rank == k_expected)
        and pairing_is_identity
    )
    report["ok"] = bool(ok)

    return report



"""QLDPC code constructions."""

from .base import QldpcCode
from .bpc import BpcCode
from .hgp import HgpCode
from .lsc import LscCode
from .qlp import QlpCode, QlpPolyCode

__all__ = ["QldpcCode", "HgpCode", "QlpCode", "QlpPolyCode", "LscCode", "BpcCode"]



"""
@author: Yingjia Lin
"""

from typing import Dict, FrozenSet, List

import numpy as np
from scipy.sparse import csc_matrix
import stim


############################Detector error matrix conversion codes##############################################################
'''
This part of the code is to convert a stim.DetectorErrorModel into a detector error matrix.
It is inspired by and adapted from:
1. BeliefMatching package (by Oscar Higgott): https://github.com/oscarhiggott/BeliefMatching/tree/main
2. Source codes for the paper "Toward Low-latency Iterative Decoding of QLDPC Codes Under Circuit-Level Noise":
    https://github.com/gongaa/SlidingWindowDecoder
    (by Anqi Gong)

'''

################################################################################################################################


def dict_to_csc_matrix_column_row(elements_dict, shape):
    '''
    Convert a dictionary into a `scipy.sparse.csc_matrix` with all the elements in this matrix are 1

    :params elements_dict: key: column indices, value: row indices
    :params shape: the shape of the resulting matrix

    :return a `scipy.sparse.csc_matrix` with column and row indices from the input dictionary. All elements are 1.
    '''
    # the non-zero elements in the matrix
    number_of_ones = sum(len(v) for v in elements_dict.values())
    data = np.ones(number_of_ones, dtype=np.uint8)
    # indices of the elements
    row_ind = np.zeros(number_of_ones, dtype=np.int64)
    col_ind = np.zeros(number_of_ones, dtype=np.int64)
    i = 0
    for col, v in elements_dict.items():
        for row in v:
            row_ind[i] = row
            col_ind[i] = col
            i += 1
    return csc_matrix((data, (row_ind, col_ind)), shape=shape)


def dict_to_csc_matrix_row_column(elements_dict, shape):
    '''
    Convert a dictionary into a `scipy.sparse.csc_matrix` with all the elements in this matrix are 1

    :params elements_dict: key: row indices, value: column indices
    :params shape: the shape of the resulting matrix

    :return a `scipy.sparse.csc_matrix` with row and column indices from the input dictionary. All elements are 1.
    '''
    # Non-zero elements of the matrix
    number_of_ones = sum(len(v) for v in elements_dict.keys())
    data = np.ones(number_of_ones, dtype=np.uint8)
    # indices of the elements
    row_ind = np.zeros(number_of_ones, dtype=np.int64)
    col_ind = np.zeros(number_of_ones, dtype=np.int64)
    i = 0
    for v, col in elements_dict.items():
        for row in v:
            row_ind[i] = row
            col_ind[i] = col
            i += 1
    return csc_matrix((data, (row_ind, col_ind)), shape=shape)


def detector_error_model_to_matrix(dem: stim.DetectorErrorModel):
    '''
    Obtain the detector error matrix from stim.DetectorErrorModel

    :param dem: Detector error model of the syndrome extraction circuit generated from stim

    :return check_matrix: detector error matrix. Each column represents a fault mechanism and each row represents a detector
    :return observables_matrix: the corresponding observable flips to each fault
    :return priors: the probability of each fault
    '''

    dem_dict: Dict[FrozenSet[int], int] = {}  # dictionary representation of detector error matrix, key: detector flips, value: fault_ids
    Logical_dict: Dict[int, FrozenSet[int]] = {}  # dictionary representation of logical observable flips, key: fault_ids, value: observable flips
    priors: List[float] = []  # error mechanism

    def handle_error(prob: float, detectors: List[int], observables: List[int]) -> None:
        dets = frozenset(detectors)
        obs = frozenset(observables)

        if dets not in dem_dict:  # for syndrome not added to the dem_dict
            dem_dict[dets] = len(dem_dict)  # key: detector flips, value: fault_ids
            priors.append(prob)  # list of probability
            Logical_dict[dem_dict[dets]] = obs  # key: fault_ids, value: observable flips
        else:
            syndrome_id = dem_dict[dets]  # get the syndrome id when the syndrome is already added into the dem_dict
            priors[syndrome_id] = priors[syndrome_id] * (1 - prob) + prob * (1 - priors[syndrome_id])  # combining the probability

    for instruction in dem.flattened():

        if instruction.type == "error":  # fault mechanism in detector error model

            dets: List[int] = []
            obs: List[int] = []
            t: stim.DemTarget
            p = instruction.args_copy()[0]
            for t in instruction.targets_copy():
                if t.is_relative_detector_id():
                    dets.append(t.val)
                elif t.is_logical_observable_id():
                    obs.append(t.val)
            if len(dets) == 0:
                print(instruction)
            handle_error(p, dets, obs)
        elif instruction.type == "detector":
            pass
        elif instruction.type == "logical_observable":
            pass
        else:
            raise NotImplementedError()
    detector_error_matrix = dict_to_csc_matrix_row_column(dem_dict,
                                      shape=(dem.num_detectors, len(dem_dict)))
    observables_matrix = dict_to_csc_matrix_column_row(Logical_dict, shape=(dem.num_observables, len(dem_dict)))
    priors = np.array(priors)
    return detector_error_matrix, observables_matrix, priors

############################################################################################################################################

############################Sliding window implementation with spacetime codes##############################################################


def spacetime(circuit, hz, W, F, num_cor_rounds):
    '''
    Obtain the spacetime slices of detector error matrix for sliding window decoder

    :param circuit: stim circuit
    :param hz: X or Z error parity check matrix of the code.
    :param W: Width of sliding window
    :param F: Width of overlap between consecutive sliding windows
    :param num_cor_rounds: number of windows before the last window

    :return window_check_set: a set of sliced spacetime detector error matrix for each window in sliding window decoder
    :return window_observable_set: a set of sliced observable matrix that marks the observable flips of each fault in each window
    :return window_priors_set: a set of probability for faults in each window
    :return window_update: the detector information update for next window of each fault mechanism in each window
    '''
    if F == 0:
        raise ValueError("Input parameter F cannot be zero.")
    model = circuit.detector_error_model(decompose_errors=False)  # detector error model of the circuit
    check_matrix, observable_matrix, priors = detector_error_model_to_matrix(model)
    window_check_set = []
    window_observable_set = []
    window_priors_set = []
    window_update = []
    col_min = 0
    '''Check_matrix for each window'''
    for k in range(num_cor_rounds):
        window_check_matrix = check_matrix[k * F * hz.shape[0]:(k * F + W) * hz.shape[0], col_min:]
        if len(window_check_matrix.indptr) == 1:
            raise ValueError("There is no noise in one of the decoding window. This means there are redundant detectors that do not check for any error.")
        col_max = np.max(np.where(np.diff(window_check_matrix.indptr) > 0)[0])  # all the columns that affect the window
        window_check_matrix = window_check_matrix[:, :col_max + 1]
        window_check_set.append(window_check_matrix)

        '''corresponding flips of observables: only care about the part we fix'''
        F_correction = window_check_matrix[:F * hz.shape[0], :]
        cor_max = np.max(np.where(np.diff(F_correction.indptr) > 0)[0])
        window_observable_matrix = observable_matrix[:, col_min:cor_max + 1 + col_min]
        window_observable_set.append(window_observable_matrix)

        '''probability of each fault'''
        window_priors = priors[col_min:col_max + 1 + col_min]
        window_priors_set.append(window_priors)
        '''updating the detector flips for the next window'''
        updated_info = check_matrix[(k + 1) * F * hz.shape[0]:((k + 1) * F + 1) * hz.shape[0], col_min:cor_max + 1 + col_min]
        col_min = (cor_max + 1) + col_min
        window_update.append(updated_info)
    '''last window check matrix'''
    last_window_check_matrix = check_matrix[F * num_cor_rounds * hz.shape[0]:, col_min:]
    window_check_set.append(last_window_check_matrix)
    '''last window observable flip'''
    last_window_observable_matrix = observable_matrix[:, col_min:]
    window_observable_set.append(last_window_observable_matrix)
    '''last window prior'''
    last_window_priors = priors[col_min:]
    window_priors_set.append(last_window_priors)

    return window_check_set, window_observable_set, window_priors_set, window_update


__all__ = [
    "dict_to_csc_matrix_column_row",
    "dict_to_csc_matrix_row_column",
    "detector_error_model_to_matrix",
    "spacetime",
]



import random

import networkx as nx
import numpy as np

from .base import CircuitBuilder


class CardinalBuilder(CircuitBuilder):
    name = "cardinal"

    def build(self, code, **opts):
        self.build_graph(code, **opts)
        return code.graph

    # Draw the Tanner graph of the code.
    def draw_graph(self, code, draw_edges=True):
        pos = nx.get_node_attributes(code.graph, 'pos')
        if not draw_edges:
            nx.draw(code.graph, pos, node_color=code.node_colors, with_labels=True, font_color='white')
            return

        edges = code.graph.edges()
        edge_colors = [code.graph[u][v]['color'] for u, v in edges]
        code.graph.add_edges_from(edges)
        nx.draw(code.graph, pos, node_color=code.node_colors, edge_color=edge_colors, with_labels=True, font_color='white')
        return

    def build_graph(self, code, **opts):
        code.graph = nx.Graph()
        code.direction_inds = {'E': 0, 'N': 1, 'S': 2, 'W': 3}
        code.direction_colors = ['green', 'blue', 'orange', 'red']

        code.node_colors = []  # 'blue' for data qubits, 'green' for zcheck qubits, 'purple' for xcheck qubits
        code.edges = [[] for _ in range(len(code.direction_inds))]  # edges of the Tanner graph of each direction

        # dictionaries used to efficiently construct the reversed Tanner graph for each direction
        code.rev_dics = [{} for _ in range(len(code.direction_inds))]
        code.rev_nodes = [[] for _ in range(len(code.direction_inds))]  # nodes of the reversed Tanner graph of each direction
        code.rev_edges = [[] for _ in range(len(code.direction_inds))]  # edges of the reversed Tanner graph of each direction.
        code.colored_edges = [{} for _ in range(len(code.direction_inds))]  # for each direction, dictionary's key is the color, values are the edges
        code.num_colors = {direction: 0 for direction in code.direction_inds.keys()}
        return

    # Helper function for assigning bool to each edge of the classical code's parity check matrix
    def get_classical_edge_bools(self, h, seed):
        c0_scores = {}
        c1_scores = {}
        edge_signs = {}
        random.seed(seed)

        for edge in np.argwhere(h == 1):
            c0, c1 = edge
            c0_score = c0_scores.get(c0, 0)
            c1_score = c1_scores.get(c1, 0)

            p = random.random()
            tf = c0_score + c1_score > 0 or (c0_score + c1_score == 0 and p >= 0.5)
            sign = int(tf) * 2 - 1
            edge_signs[(c0, c1)] = tf
            c0_scores[c0] = c0_scores.get(c0, 0) - sign
            c1_scores[c1] = c1_scores.get(c1, 0) - sign

        return edge_signs

    # Helper function for adding edges
    def add_edge(self, code, edge_no, direction_ind, control, target):
        code.edges[direction_ind] += [(control, target)]
        code.graph.add_edge(control, target, color=code.direction_colors[direction_ind])

        # add edge to rev graph
        code.rev_nodes[direction_ind] += [edge_no]
        if control not in code.rev_dics[direction_ind]:
            code.rev_dics[direction_ind][control] = [edge_no]
        else:
            code.rev_dics[direction_ind][control] += [edge_no]
        if target not in code.rev_dics[direction_ind]:
            code.rev_dics[direction_ind][target] = [edge_no]
        else:
            code.rev_dics[direction_ind][target] += [edge_no]
        return

    def color_edges(self, code):
        # Construct the reversed Tanner graph's edges from rev_dics dictionary
        for direction_ind in range(len(code.rev_edges)):
            dic = code.rev_dics[direction_ind]
            for nodes in dic.values():
                for i in range(len(nodes) - 1):
                    for j in range(i + 1, len(nodes)):
                        code.rev_edges[direction_ind] += [(nodes[i], nodes[j])]

        # list of colors of the reversed Tanner graph's nodes for each direction
        edge_colors = [[] for _ in range(len(code.direction_inds))]
        # Apply coloring to the reversed Tanner graph
        for direction_ind in range(len(code.rev_edges)):
            rev_graph = nx.Graph()
            rev_graph.add_nodes_from(code.rev_nodes[direction_ind])
            rev_graph.add_edges_from(code.rev_edges[direction_ind])

            edge_coloration = nx.greedy_color(rev_graph)
            # Somehow the dictionary returned by nx.greedy_color shuffles the keys (rev_nodes[direction_ind])
            # so the values (colors) need to be shuffled correctly.
            paired = list(zip(edge_coloration.keys(), edge_coloration.values()))
            paired_sorted = sorted(paired, key=lambda x: x[0])
            _, reordered_colors = zip(*paired_sorted)
            edge_colors[direction_ind] = reordered_colors

        # Construct colored_edges (dictionary of edges of each direction and color)
        for direction_ind in range(len(code.colored_edges)):
            for i in range(len(code.edges[direction_ind])):
                edge = list(code.edges[direction_ind][i])
                color = edge_colors[direction_ind][i]

                if color not in code.colored_edges[direction_ind]:
                    code.colored_edges[direction_ind][color] = edge
                else:
                    code.colored_edges[direction_ind][color] += edge

        for direction in list(code.direction_inds.keys()):
            direction_ind = code.direction_inds[direction]
            code.num_colors[direction] = len(list(code.colored_edges[direction_ind].keys()))
        return



"""Public API facade for QUITS."""

from .circuit import get_qldpc_mem_circuit
from .decoder import (
    detector_error_model_to_matrix,
    sliding_window_bplsd_circuit_mem,
    sliding_window_bplsd_phenom_mem,
    sliding_window_bposd_circuit_mem,
    sliding_window_bposd_phenom_mem,
    sliding_window_circuit_mem,
    sliding_window_phenom_mem,
)
from .qldpc_code import BpcCode, HgpCode, LscCode, QldpcCode, QlpCode, QlpPolyCode
from .simulation import get_stim_mem_result

__all__ = [
    "BpcCode",
    "HgpCode",
    "QldpcCode",
    "QlpCode",
    "QlpPolyCode",
    "LscCode",
    "get_qldpc_mem_circuit",
    "get_stim_mem_result",
    "detector_error_model_to_matrix",
    "sliding_window_phenom_mem",
    "sliding_window_bposd_phenom_mem",
    "sliding_window_bplsd_phenom_mem",
    "sliding_window_circuit_mem",
    "sliding_window_bposd_circuit_mem",
    "sliding_window_bplsd_circuit_mem",
]



"""
@author: Mingyu Kang
"""

import numpy as np
from tqdm import tqdm

def get_stim_mem_result(circuit, num_trials, seed=-1):
    '''
    Simulates the Stim circuit and obtains the detector and logical observable flip results for logical memory. 

    :param circuit: Stim circuit
    :param num_trials: Number of trials to sample from
    :param seed: Random seed

    :return detections: 2-dim numpy array of shape (num_trials, # detectors). Each entry is a bit (0 or 1) that represents the detector result. 
                        If get_all_detections, # detectors = num_(basis)check * (num_rounds+2) 
                        Else, # detectors = num_(basis)check * (num_rounds+2) + num_(other basis)check * num_rounds
    :return observable_flips: 2-dim numpy array of shape (num_trials, number of logical qubits). Each entry is a bit (0 or 1) that represents whether the logical Z flipped.
    '''

    if seed >= 0: 
        sampler = circuit.compile_detector_sampler(seed=seed)
    else:
        sampler = circuit.compile_detector_sampler()
    
    detection_events, observable_flips = sampler.sample(shots=num_trials, separate_observables=True)
    return detection_events, observable_flips


def get_codecap_pL(code, p, num_trials, decoder, dict, basis='Z', seed=-1, tqdm_on=False):

    if seed >= 0:
        np.random.seed(seed)
    if tqdm_on:
        iterator = tqdm(range(num_trials))
    else:
        iterator = range(num_trials)

    parity_check_matrix = code.hz if basis == 'Z' else code.hx
    logical_codewords = code.lz if basis == 'Z' else code.lx

    bpd = decoder(parity_check_matrix, **dict)

    num_errors=0
    for i in iterator:
        noise = np.random.binomial(1, p, parity_check_matrix.shape[1])
        syndrome= parity_check_matrix @ noise % 2
        decoded_error = bpd.decode(syndrome)
        residual_error = (decoded_error + noise) %2
        if (logical_codewords @ residual_error % 2).any():
            num_errors += 1
    
    return num_errors/num_trials


"""
@author: Yingjia Lin
"""

from ldpc.bplsd_decoder import BpLsdDecoder

from .sliding_window import sliding_window_phenom_mem, sliding_window_circuit_mem


def sliding_window_bplsd_phenom_mem(zcheck_samples, hz, lz, W, F, eff_error_rate_per_fault: float = None, max_iter=2, lsd_order=0, bp_method='product_sum', schedule='serial', lsd_method='lsd_cs', tqdm_on=False, error_rate: float = None):
    '''
    Sliding window decoder in S. Huang and S. Puri, PRA 110, 012453 (2024) implemented with BP-LSD decoder
    For convenience the notation assumes z-type memory, but the code works equivalently for x-type memory.

    :param zcheck_samples: 2-dim numpy array of detector results; see get_stim_Zmem_result in simulation.py. Shape (# trials, # Z-check qubits * (# rounds+1))
    :param hz: Parity check matrix (in the code-capacity level) representing the Z stabilizers of the qec code. Shape ((# Z-check qubits, # data qubits))
    :param lz: Logical codeword matrix of the qec code. Shape ((# logical qubits, # data qubits))
    :param W: Width of sliding window
    :param F: Width of overlap between consecutive sliding windows
    :param eff_error_rate_per_fault: Estimate of error rate in the context of code-capacity/phenomenological level simulations.
    :param error_rate: Deprecated alias for eff_error_rate_per_fault.
                For circuit level, we suggest p * (num_layers + 3), where p is the depolarizing error rate
                and num_layers is the circuit depth for each stabilizer measurement round
                e.g. for hgp codes, num_layers == code.count_color('east') + code.count_color('north') + code.count_color('south') + code.count_color('west')
    :param max_iter: Maximum number of iterations for BP
    :param lsd_order: Lsd search depth
    :param bp_method: BP method for BP_LSD. Choose from ‘product_sum’ or ‘minimum_sum’
    :param schedule: choose from 'serial' or 'parallel'
    :param lsd_method: choose from:  'lsd_e', 'lsd_cs', 'lsd_0'

    :return logical_z_pred: Decoder's prediction of whether the logical Z codewords flipped. Shape (# trials, # logical quits)
    '''
    if eff_error_rate_per_fault is None:
        eff_error_rate_per_fault = error_rate
    if eff_error_rate_per_fault is None:
        raise ValueError("eff_error_rate_per_fault must be provided (or use deprecated error_rate).")
    # parameters of decoders
    dict1 = {'bp_method': bp_method,
            'max_iter': max_iter,
            'schedule': schedule,
            'lsd_method': lsd_method,
            'lsd_order': lsd_order,
          'error_rate': float(eff_error_rate_per_fault)}
    dict2 = {'bp_method': bp_method,
            'max_iter': max_iter,
            'schedule': schedule,
            'lsd_method': lsd_method,
            'lsd_order': lsd_order,
          'error_rate': float(eff_error_rate_per_fault)}
    logical_pred = sliding_window_phenom_mem(zcheck_samples, hz, lz, W, F, BpLsdDecoder, BpLsdDecoder, dict1, dict2, 'decode', 'decode', tqdm_on=tqdm_on)
    return logical_pred


def sliding_window_bplsd_circuit_mem(zcheck_samples, circuit, hz, lz, W, F, max_iter=2, lsd_order=0, bp_method='product_sum', schedule='serial', lsd_method='lsd_cs', tqdm_on=False):
    '''
    Sliding window decoder in S. Huang and S. Puri, PRA 110, 012453 (2024) implemented with BP-LSD decoder and spacetime detector error matrix
    For convenience the notation assumes z-type memory, but the code works equivalently for x-type memory.

    :param zcheck_samples: 2-dim numpy array of detector results; see get_stim_Zmem_result in simulation.py. Shape (# trials, # Z-check qubits * (# rounds+1))
    :param circuit: syndrome extraction circuit
    :param hz: Parity check matrix (in the code-capacity level) representing the Z stabilizers of the qec code. Shape ((# Z-check qubits, # data qubits))
    :param lz: Logical codeword matrix of the qec code. Shape ((# logical qubits, # data qubits))
    :param W: Width of sliding window
    :param F: Width of overlap between consecutive sliding windows
    :param max_iter: Maximum number of iterations for BP
    :param lsd_order: Lsd search depth
    :param bp_method: BP method for BP_LSD. Choose from ‘product_sum’ or ‘minimum_sum’
    :param schedule: choose from 'serial' or 'parallel'
    :param lsd_method: choose from:  'lsd_e', 'lsd_cs', 'lsd_0'

    :return logical_z_pred: Decoder's prediction of whether the logical Z codewords flipped. Shape (# trials, # logical quits)
    '''
    # parameters of decoders
    dict1 = {'bp_method': bp_method,
            'max_iter': max_iter,
            'schedule': schedule,
            'lsd_method': lsd_method,
            'lsd_order': lsd_order}
    dict2 = {'bp_method': bp_method,
            'max_iter': max_iter,
            'schedule': schedule,
            'lsd_method': lsd_method,
            'lsd_order': lsd_order}
    logical_pred = sliding_window_circuit_mem(zcheck_samples, circuit, hz, lz, W, F, BpLsdDecoder, BpLsdDecoder, dict1, dict2, 'channel_probs', 'channel_probs', 'decode', 'decode', tqdm_on=tqdm_on)

    return logical_pred


__all__ = ["sliding_window_bplsd_phenom_mem", "sliding_window_bplsd_circuit_mem"]
